/**************************************************************************
* DSemu: Scanline-based tiled/bitmap graphics (gpu.c)                     *
* Released under the terms of the BSD Public Licence                      *
* Imran Nazar (tf@oopsilon.com), 2004                                     *
**************************************************************************/

#include <windows.h>
#include <stdio.h>
#include "err.h"
#include "defs.h"
#include "ioreg.h"
#include "gpu.h"
#include "dma.h"
#include "vtbl.h"
#include "int.h"
#include "res.h"

u16 *VRAM;
u8  *VRAM8;

u16 BGPAL[256];
u16 OBJPAL[256];
u8 *BGPAL8=(u8*)BGPAL;
u8 *OBJPAL8=(u8*)OBJPAL;

u8 curline;

int GPUinit()
{
    u8 r,g,b; char str[50]; int a;

    VRAM = (u16*)malloc(96*1024);
    if(!VRAM) RETFAIL("GPU: FAIL: VRAM allocation.");
    VRAM8=(u8*)VRAM;

    GPUreset();

    RETPASS("GPU: Initialised.");
}

void GPUfini()
{
    free(VRAM);
    logvt->append("GPU: Shutdown.");
}

void GPUreset()
{
    int a;
    memset(VRAM, 0, 96*1024);
    memset(BGPAL,0, 256*2);
    memset(OBJPAL,0, 256*2);
    curline=0;
}

#define GPUscanMode0() \
    voff=curline*240; \
    scrnbase=(((GBAio[REG_BG0CNT].data&0x1F00)+ \
             ((curline+GBAio[REG_BG0VOFS].data)&0xF8))<<2)+ \
             ((GBAio[REG_BG0HOFS].data&0xF8)>>3); \
    scrnbaseorig=scrnbase&0xFFE0; \
    charbase=((GBAio[REG_BG0CNT].data&0x000C)<<14); \
    if(VRAM[scrnbase]&0x0800) charline=(7-((curline+GBAio[REG_BG0VOFS].data))); \
    else charline=((curline+GBAio[REG_BG0VOFS].data)); \
    if(GBAio[REG_BG0CNT].data&0x0080) \
    { \
        soff=GBAio[REG_BG0HOFS].data&7; \
        charoff=(VRAM[scrnbase]&1023)*64+(charline&7)*8; \
        for(a=0;a<240;a++) \
        { \
            if(VRAM[scrnbase]&0x0400) \
                screen[voff+a]=BGPAL[VRAM8[charoff+7-soff]]; \
            else \
                screen[voff+a]=BGPAL[VRAM8[charoff+soff]]; \
	    soff++; if(soff==8) \
	    { \
                scrnbase++; \
                if((scrnbase&0xFFE0)!=scrnbaseorig) scrnbase=scrnbaseorig; \
                charoff=(VRAM[scrnbase]&1023)*64+(charline&7)*8; \
                soff=0; \
	    } \
        } \
    } else { \
        soff=GBAio[REG_BG0HOFS].data&7; \
        charoff=(VRAM[scrnbase]&1023)*32+(charline&7)*4; \
        palnum=(VRAM[scrnbase]&0xF000)>>8; \
        for(a=0;a<240;a++) \
        { \
            if(VRAM[scrnbase]&0x0400) \
            { \
                if(a&1) \
		    screen[voff+a]= \
   	                BGPAL[palnum+(VRAM8[charoff+3-(soff>>1)]&15)]; \
                else \
		    screen[voff+a]= \
   	                BGPAL[palnum+(VRAM8[charoff+3-(soff>>1)]>>4)]; \
            } else { \
                if(a&1) \
		    screen[voff+a]= \
   	                BGPAL[palnum+(VRAM8[charoff+(soff>>1)]&15)]; \
                else \
		    screen[voff+a]= \
   	                BGPAL[palnum+(VRAM8[charoff+(soff>>1)]>>4)]; \
            } \
	    soff++; \
	    if(soff==8) \
	    { \
                scrnbase++; \
                if((scrnbase&0xFFE0)!=scrnbaseorig) scrnbase=scrnbaseorig; \
                charoff=(VRAM[scrnbase]&1023)*32+(charline&7)*4; \
                soff=0; \
	    } \
        } \
    }

#define GPUscanMode3() \
    voff=curline*240; \
    /*for(b=0;b<240;b++) screen[voff+b]=VRAM[voff+b]*/ \
    memcpy(screen+voff, VRAM+voff, 480);

#define GPUscanMode4() \
    voff=curline*240; \
    base=((GBAio[REG_DISPCNT].data&0x0010)?0xA000:0)+voff; \
    for(b=0;b<240;b++) \
        screen[voff+b]=BGPAL[VRAM8[base+/*voff+*/b]]

#define GPUscanMode5() \
    soff=curline*240; \
    memset(screen+soff,0,480); \
    if(curline>=16 && curline<144) \
    { \
        /*voff=(curline-16)*160;*/ \
        base=((GBAio[REG_DISPCNT].data&0x0010)?0x5000:0)+(curline-16)*160; \
	/*for(b=0;b<40;b++) \
	{ \
	    screen[soff+b]=0; \
	    screen[soff+b+200]=0; \
	} \
        for(b=0;b<160;b++) \
            screen[soff+b+40]= \
	        VRAM[base+voff+b];*/ \
	memcpy(screen+soff+40,VRAM+base/*+voff*/,320); \
    } \
    /*else /*for(b=0;b<240;b++) screen[soff+b]=0*/ \
        /*memset(screen+soff,0,480)*/

void GPUscanline(u16 *screen)
{
    int x,y,a,b;
    u32 base, voff, soff;
    u32 scrnbase, scrnbaseorig, charbase, charline, charoff, palnum;

    if(curline<160)
    {
        switch(GBAio[REG_DISPCNT].data&7)
        {
            case 0: GPUscanMode0(); break;
            case 3: GPUscanMode3(); break;
            case 4: GPUscanMode4(); break;
            case 5: GPUscanMode5(); break;
        }
        if(GBAio[REG_DISPSTAT].data&STAT_INTHBL) IntFire(INT_HBLANK);
        DMAcheck(DMA_TIMEHBL);
    }
    GBAio[REG_DISPSTAT].data|=STAT_HBLANK;
}

void GPUclearHBL()
{
    GBAio[REG_DISPSTAT].data&=(0xFFFF-STAT_HBLANK);
    curline++; if(curline>227) curline=0;
    GBAio[REG_VCOUNT].data=curline;
    switch(curline)
    {
        case 0:   GBAio[REG_DISPSTAT].data&=(0xFFFF-STAT_VBLANK); break;
        case 160:
	   GBAio[REG_DISPSTAT].data|=STAT_VBLANK;
           if(GBAio[REG_DISPSTAT].data&STAT_INTVBL) IntFire(INT_VBLANK);
           DMAcheck(DMA_TIMEVBL);
	   break;
    }
    if(curline==(GBAio[REG_DISPSTAT].data>>6))
    {
        GBAio[REG_DISPSTAT].data|=STAT_VCOUNT;
        if(GBAio[REG_DISPSTAT].data&STAT_INTVCT) IntFire(INT_VCOUNT);
    }
    else
        GBAio[REG_DISPSTAT].data&=(0xFFFF-STAT_VCOUNT);
}

void GPUpalcol(u16 *pbuf, int idx)
{
    char str[80];
    u16 col=BGPAL[idx];
    u8 r=(col&0x001F), g=(col&0x03E0)>>5, b=(col&0x7C00)>>10;
    dbgOutClear(pbuf+128*136,128,8);
    sprintf(str,"R:%02Xh  G:%02Xh  B:%02Xh",r,g,b);
    dbgOut(pbuf+128*136,128,8,str,0,0,0x7FFF);
}

void GPUpal(u16 *pbuf)
{
    int a,b,c,d; u32 xoff=0,yoff=0,idx=0;
    dbgOut(pbuf,128,144,"BGPAL",0,0,0x7FFF);
    for(a=0;a<16;a++)
    {
        for(b=0;b<16;b++)
        {
	    for(c=0;c<8;c++)
	    {
	        for(d=0;d<8;d++)
	        {
//	            pbuf[(143-(a*8+c))*128+(b*8+d)]=LCDcolour[BGPAL[a*16+b]];
//                  pbuf[off]=LCDcolour[BGPAL[idx]];
//                  off++;
	            pbuf[yoff*128+xoff]=BGPAL[idx];
                    xoff++;
	        }
//	        off+=137;
                yoff++; xoff-=8;
	    }
//	    off-=1008;
	    idx++;
            xoff+=8; yoff-=8;
	}
//	idx+=16;
	xoff=0; yoff+=8;
    }
}

void GPUwrH(u32 addr, u16 val)
{
    char str[80]; u32 reg;
    switch(addr&0x0F000000)
    {
        case 0x04000000:
            reg=(addr&0x5F)>>1;
	    if(GBAio[reg].flags&REG_FLAG_W)
	        GBAio[reg].data=val;
//	    sprintf(str,"GPU: write %08X=%04X",addr,val);
//	    logvt->append(str);
	    break;
        case 0x05000000:
	    if(addr&0x00000200) OBJPAL[(addr&0x1FF)>>1]=val;
	    else                BGPAL [(addr&0x1FF)>>1]=val;
	    break;
        case 0x06000000:
//	    sprintf(str,"GPU: write %08X=%04X",addr,val);
//	    logvt->append(str);
	    VRAM[(addr&0x0001FFFF)>>1]=val; break;
    }
}

void GPUwrW(u32 addr, u32 val)
{
    GPUwrH(addr, val&65535);
    GPUwrH(addr+2, val>>16);
}

u16 GPUrdH(u32 addr)
{
    char str[80]; u32 reg;
    switch(addr&0x0F000000)
    {
        case 0x04000000:
            reg=(addr&0x5F)>>1;
//	    sprintf(str,"GPU: read %08X=%04X",addr,regs[reg].data);
//	    logvt->append(str);
	    if(GBAio[reg].flags&REG_FLAG_R)
	        return GBAio[reg].data;
        case 0x05000000:
            if(addr&0x00000200) return OBJPAL[(addr&0x1FF)>>1];
            else                return BGPAL [(addr&0x1FF)>>1];
        case 0x06000000:
//	    sprintf(str,"GPU: read %08X=%04X",addr,VRAM[(addr&0x0001FFFF)>>1]);
//	    logvt->append(str);
	    return VRAM[(addr&0x0001FFFF)>>1];
    }
}

u32 GPUrdW(u32 addr)
{
    u16 vall,valh; u32 val;
    vall=GPUrdH(addr);
    valh=GPUrdH(addr+2);
    val=(valh<<16)+vall;
    return val;
}

void GPUstatus()
{
    char str[512];
}

/*** EOF:gpu.c ***********************************************************/

