/**************************************************************************
* DSemu: ARM7 core definitions and opcode pointer table (arm7func.h)      *
* Released under the terms of the BSD Public Licence                      *
* Imran Nazar (tf@oopsilon.com), 2004                                     *
**************************************************************************/

#ifndef __ARM7FUNC_H_
#define __ARM7FUNC_H_

#include "defs.h"
#include "cache.h"

int ARM7exec(), ARM7execfor(u32);
void ARM7dbgOut();
void ARM7modesw(int,int);
void ARM7status(int,int);
void ARM7updateCPSR(), ARM7splitCPSR();
int ARM7init(char*,u32*);
void ARM7fini();
void ARM7reset();
__int64 ARM7getClock();
void ARM7addClock(int);
void ARM7irq();

extern int MMUinit();
extern void MMUfini();
extern void MMUreg(int,void*);
extern void MMUunreg(int);
extern u8  MMUrdB(u32, u32);
extern u16 MMUrdH(u32, u32);
extern u32 MMUrdW(u32, u32);
extern RAMWORD MMUrdS(u32, u32);
extern void MMUwrB(u32, u32, u8 );
extern void MMUwrH(u32, u32, u16);
extern void MMUwrW(u32, u32, u32);
extern void MMUwrS(u32, u32, RAMWORD);

extern char *ARM7DASM(u32);

char ARMmodes[7][4]={"USR","SYS","FIQ","IRQ","SUP","ABT","UND"};
int ARMmodetocpsr[7]={ARMS_M_USR,ARMS_M_SYS,ARMS_M_FIQ,ARMS_M_IRQ,
                      ARMS_M_SVC,ARMS_M_ABT,ARMS_M_UND};
int ARMcpsrtomode[16]={0,2,3,4,-1,-1,-1,5,-1,-1,-1,6,-1,-1,-1,1};

int ARM7condEQ(), ARM7condNE(), ARM7condCS(), ARM7condCC();
int ARM7condMI(), ARM7condPL(), ARM7condVS(), ARM7condVC();
int ARM7condHI(), ARM7condLS(), ARM7condGE(), ARM7condLT();
int ARM7condGT(), ARM7condLE(), ARM7condAL(), ARM7condNV();

static intfptr ARM7cond[]={
  ARM7condEQ, ARM7condNE, ARM7condCS, ARM7condCC,
  ARM7condMI, ARM7condPL, ARM7condVS, ARM7condVC,
  ARM7condHI, ARM7condLS, ARM7condGE, ARM7condLT,
  ARM7condGT, ARM7condLE, ARM7condAL, ARM7condNV,
};

//---UNdefined/predictable-------------------------------------------------

void ARM7opUNP();
void ARM7opUND();
void ARM7opUNI();
void ARM7opUNL();

//---Branch (with Link)----------------------------------------------------

void ARM7opB(), ARM7opBL();

void ARM7opBX();
void ARM7opBLXr();

//---DP (DataProcessing); LsL/LsR/AsR/RoR; Immediate/Register--------------

void ARM7opANDlli(), ARM7opANDllr();
void ARM7opANDlri(), ARM7opANDlrr();
void ARM7opANDari(), ARM7opANDarr();
void ARM7opANDrri(), ARM7opANDrrr();
void ARM7opANDimm();

void ARM7opEORlli(), ARM7opEORllr();
void ARM7opEORlri(), ARM7opEORlrr();
void ARM7opEORari(), ARM7opEORarr();
void ARM7opEORrri(), ARM7opEORrrr();
void ARM7opEORimm();

void ARM7opSUBlli(), ARM7opSUBllr();
void ARM7opSUBlri(), ARM7opSUBlrr();
void ARM7opSUBari(), ARM7opSUBarr();
void ARM7opSUBrri(), ARM7opSUBrrr();
void ARM7opSUBimm();

void ARM7opRSBlli(), ARM7opRSBllr();
void ARM7opRSBlri(), ARM7opRSBlrr();
void ARM7opRSBari(), ARM7opRSBarr();
void ARM7opRSBrri(), ARM7opRSBrrr();
void ARM7opRSBimm();

void ARM7opADDlli(), ARM7opADDllr();
void ARM7opADDlri(), ARM7opADDlrr();
void ARM7opADDari(), ARM7opADDarr();
void ARM7opADDrri(), ARM7opADDrrr();
void ARM7opADDimm();

void ARM7opADClli(), ARM7opADCllr();
void ARM7opADClri(), ARM7opADClrr();
void ARM7opADCari(), ARM7opADCarr();
void ARM7opADCrri(), ARM7opADCrrr();
void ARM7opADCimm();

void ARM7opSBClli(), ARM7opSBCllr();
void ARM7opSBClri(), ARM7opSBClrr();
void ARM7opSBCari(), ARM7opSBCarr();
void ARM7opSBCrri(), ARM7opSBCrrr();
void ARM7opSBCimm();

void ARM7opRSClli(), ARM7opRSCllr();
void ARM7opRSClri(), ARM7opRSClrr();
void ARM7opRSCari(), ARM7opRSCarr();
void ARM7opRSCrri(), ARM7opRSCrrr();
void ARM7opRSCimm();

void ARM7opORRlli(), ARM7opORRllr();
void ARM7opORRlri(), ARM7opORRlrr();
void ARM7opORRari(), ARM7opORRarr();
void ARM7opORRrri(), ARM7opORRrrr();
void ARM7opORRimm();

void ARM7opMOVlli(), ARM7opMOVllr();
void ARM7opMOVlri(), ARM7opMOVlrr();
void ARM7opMOVari(), ARM7opMOVarr();
void ARM7opMOVrri(), ARM7opMOVrrr();
void ARM7opMOVimm();

void ARM7opBIClli(), ARM7opBICllr();
void ARM7opBIClri(), ARM7opBIClrr();
void ARM7opBICari(), ARM7opBICarr();
void ARM7opBICrri(), ARM7opBICrrr();
void ARM7opBICimm();

void ARM7opMVNlli(), ARM7opMVNllr();
void ARM7opMVNlri(), ARM7opMVNlrr();
void ARM7opMVNari(), ARM7opMVNarr();
void ARM7opMVNrri(), ARM7opMVNrrr();
void ARM7opMVNimm();

void ARM7opANDSlli(), ARM7opANDSllr();
void ARM7opANDSlri(), ARM7opANDSlrr();
void ARM7opANDSari(), ARM7opANDSarr();
void ARM7opANDSrri(), ARM7opANDSrrr();
void ARM7opANDSimm();

void ARM7opEORSlli(), ARM7opEORSllr();
void ARM7opEORSlri(), ARM7opEORSlrr();
void ARM7opEORSari(), ARM7opEORSarr();
void ARM7opEORSrri(), ARM7opEORSrrr();
void ARM7opEORSimm();

void ARM7opSUBSlli(), ARM7opSUBSllr();
void ARM7opSUBSlri(), ARM7opSUBSlrr();
void ARM7opSUBSari(), ARM7opSUBSarr();
void ARM7opSUBSrri(), ARM7opSUBSrrr();
void ARM7opSUBSimm();

void ARM7opRSBSlli(), ARM7opRSBSllr();
void ARM7opRSBSlri(), ARM7opRSBSlrr();
void ARM7opRSBSari(), ARM7opRSBSarr();
void ARM7opRSBSrri(), ARM7opRSBSrrr();
void ARM7opRSBSimm();

void ARM7opADDSlli(), ARM7opADDSllr();
void ARM7opADDSlri(), ARM7opADDSlrr();
void ARM7opADDSari(), ARM7opADDSarr();
void ARM7opADDSrri(), ARM7opADDSrrr();
void ARM7opADDSimm();

void ARM7opADCSlli(), ARM7opADCSllr();
void ARM7opADCSlri(), ARM7opADCSlrr();
void ARM7opADCSari(), ARM7opADCSarr();
void ARM7opADCSrri(), ARM7opADCSrrr();
void ARM7opADCSimm();

void ARM7opSBCSlli(), ARM7opSBCSllr();
void ARM7opSBCSlri(), ARM7opSBCSlrr();
void ARM7opSBCSari(), ARM7opSBCSarr();
void ARM7opSBCSrri(), ARM7opSBCSrrr();
void ARM7opSBCSimm();

void ARM7opRSCSlli(), ARM7opRSCSllr();
void ARM7opRSCSlri(), ARM7opRSCSlrr();
void ARM7opRSCSari(), ARM7opRSCSarr();
void ARM7opRSCSrri(), ARM7opRSCSrrr();
void ARM7opRSCSimm();

void ARM7opTSTSlli(), ARM7opTSTSllr();
void ARM7opTSTSlri(), ARM7opTSTSlrr();
void ARM7opTSTSari(), ARM7opTSTSarr();
void ARM7opTSTSrri(), ARM7opTSTSrrr();
void ARM7opTSTSimm();

void ARM7opTEQSlli(), ARM7opTEQSllr();
void ARM7opTEQSlri(), ARM7opTEQSlrr();
void ARM7opTEQSari(), ARM7opTEQSarr();
void ARM7opTEQSrri(), ARM7opTEQSrrr();
void ARM7opTEQSimm();

void ARM7opCMPSlli(), ARM7opCMPSllr();
void ARM7opCMPSlri(), ARM7opCMPSlrr();
void ARM7opCMPSari(), ARM7opCMPSarr();
void ARM7opCMPSrri(), ARM7opCMPSrrr();
void ARM7opCMPSimm();

void ARM7opCMNSlli(), ARM7opCMNSllr();
void ARM7opCMNSlri(), ARM7opCMNSlrr();
void ARM7opCMNSari(), ARM7opCMNSarr();
void ARM7opCMNSrri(), ARM7opCMNSrrr();
void ARM7opCMNSimm();

void ARM7opORRSlli(), ARM7opORRSllr();
void ARM7opORRSlri(), ARM7opORRSlrr();
void ARM7opORRSari(), ARM7opORRSarr();
void ARM7opORRSrri(), ARM7opORRSrrr();
void ARM7opORRSimm();

void ARM7opMOVSlli(), ARM7opMOVSllr();
void ARM7opMOVSlri(), ARM7opMOVSlrr();
void ARM7opMOVSari(), ARM7opMOVSarr();
void ARM7opMOVSrri(), ARM7opMOVSrrr();
void ARM7opMOVSimm();

void ARM7opBICSlli(), ARM7opBICSllr();
void ARM7opBICSlri(), ARM7opBICSlrr();
void ARM7opBICSari(), ARM7opBICSarr();
void ARM7opBICSrri(), ARM7opBICSrrr();
void ARM7opBICSimm();

void ARM7opMVNSlli(), ARM7opMVNSllr();
void ARM7opMVNSlri(), ARM7opMVNSlrr();
void ARM7opMVNSari(), ARM7opMVNSarr();
void ARM7opMVNSrri(), ARM7opMVNSrrr();
void ARM7opMVNSimm();

//---AR (ARithmetic); REGister/IMMediate-----------------------------------

void ARM7opMUL(), ARM7opMLA();
void ARM7opUMULL(), ARM7opUMLAL(), ARM7opSMULL(), ARM7opSMLAL();

void ARM7opMULS(), ARM7opMLAS();
void ARM7opUMULLS(), ARM7opUMLALS(), ARM7opSMULLS(), ARM7opSMLALS();

//---Move Register From Status; Imm/Reg; Cpsr/Spsr-------------------------

void ARM7opMRSrc(), ARM7opMRSrs();

//---Move Status from Register; Cpsr/Spsr----------------------------------

void ARM7opMSRrc(), ARM7opMSRrs();
void ARM7opMSRic(), ARM7opMSRis();

//---LS (Load/Store); PReidx/OFfset/PsTidx; Imm/Reg; Minus/Plus; (Shift)---

void ARM7opLDRptim(), ARM7opLDRofim(), ARM7opLDRprim();
void ARM7opLDRptip(), ARM7opLDRofip(), ARM7opLDRprip();
void ARM7opLDRptrm(), ARM7opLDRofrm(), ARM7opLDRprrm();
void ARM7opLDRptrp(), ARM7opLDRofrp(), ARM7opLDRprrp();

void ARM7opSTRptim(), ARM7opSTRofim(), ARM7opSTRprim();
void ARM7opSTRptip(), ARM7opSTRofip(), ARM7opSTRprip();
void ARM7opSTRptrm(), ARM7opSTRofrm(), ARM7opSTRprrm();
void ARM7opSTRptrp(), ARM7opSTRofrp(), ARM7opSTRprrp();

void ARM7opLDRBptim(), ARM7opLDRBofim(), ARM7opLDRBprim();
void ARM7opLDRBptip(), ARM7opLDRBofip(), ARM7opLDRBprip();
void ARM7opLDRBptrm(), ARM7opLDRBofrm(), ARM7opLDRBprrm();
void ARM7opLDRBptrp(), ARM7opLDRBofrp(), ARM7opLDRBprrp();

void ARM7opSTRBptim(), ARM7opSTRBofim(), ARM7opSTRBprim();
void ARM7opSTRBptip(), ARM7opSTRBofip(), ARM7opSTRBprip();
void ARM7opSTRBptrm(), ARM7opSTRBofrm(), ARM7opSTRBprrm();
void ARM7opSTRBptrp(), ARM7opSTRBofrp(), ARM7opSTRBprrp();

void ARM7opLDRHptim(), ARM7opLDRHofim(), ARM7opLDRHprim();
void ARM7opLDRHptip(), ARM7opLDRHofip(), ARM7opLDRHprip();
void ARM7opLDRHptrm(), ARM7opLDRHofrm(), ARM7opLDRHprrm();
void ARM7opLDRHptrp(), ARM7opLDRHofrp(), ARM7opLDRHprrp();

void ARM7opSTRHptim(), ARM7opSTRHofim(), ARM7opSTRHprim();
void ARM7opSTRHptip(), ARM7opSTRHofip(), ARM7opSTRHprip();
void ARM7opSTRHptrm(), ARM7opSTRHofrm(), ARM7opSTRHprrm();
void ARM7opSTRHptrp(), ARM7opSTRHofrp(), ARM7opSTRHprrp();

void ARM7opLDRTptim(), ARM7opLDRTofim(), ARM7opLDRTprim();
void ARM7opLDRTptip(), ARM7opLDRTofip(), ARM7opLDRTprip();
void ARM7opLDRTptrm(), ARM7opLDRTofrm(), ARM7opLDRTprrm();
void ARM7opLDRTptrp(), ARM7opLDRTofrp(), ARM7opLDRTprrp();

void ARM7opSTRTptim(), ARM7opSTRTofim(), ARM7opSTRTprim();
void ARM7opSTRTptip(), ARM7opSTRTofip(), ARM7opSTRTprip();
void ARM7opSTRTptrm(), ARM7opSTRTofrm(), ARM7opSTRTprrm();
void ARM7opSTRTptrp(), ARM7opSTRTofrp(), ARM7opSTRTprrp();

void ARM7opLDRBTptim(), ARM7opLDRBTofim(), ARM7opLDRBTprim();
void ARM7opLDRBTptip(), ARM7opLDRBTofip(), ARM7opLDRBTprip();
void ARM7opLDRBTptrm(), ARM7opLDRBTofrm(), ARM7opLDRBTprrm();
void ARM7opLDRBTptrp(), ARM7opLDRBTofrp(), ARM7opLDRBTprrp();

void ARM7opSTRBTptim(), ARM7opSTRBTofim(), ARM7opSTRBTprim();
void ARM7opSTRBTptip(), ARM7opSTRBTofip(), ARM7opSTRBTprip();
void ARM7opSTRBTptrm(), ARM7opSTRBTofrm(), ARM7opSTRBTprrm();
void ARM7opSTRBTptrp(), ARM7opSTRBTofrp(), ARM7opSTRBTprrp();

void ARM7opLDRSHptim(), ARM7opLDRSHofim(), ARM7opLDRSHprim();
void ARM7opLDRSHptip(), ARM7opLDRSHofip(), ARM7opLDRSHprip();
void ARM7opLDRSHptrm(), ARM7opLDRSHofrm(), ARM7opLDRSHprrm();
void ARM7opLDRSHptrp(), ARM7opLDRSHofrp(), ARM7opLDRSHprrp();

void ARM7opLDRSBptim(), ARM7opLDRSBofim(), ARM7opLDRSBprim();
void ARM7opLDRSBptip(), ARM7opLDRSBofip(), ARM7opLDRSBprip();
void ARM7opLDRSBptrm(), ARM7opLDRSBofrm(), ARM7opLDRSBprrm();
void ARM7opLDRSBptrp(), ARM7opLDRSBofrp(), ARM7opLDRSBprrp();

void ARM7opLDRptrmll(), ARM7opLDRofrmll(), ARM7opLDRprrmll();
void ARM7opLDRptrpll(), ARM7opLDRofrpll(), ARM7opLDRprrpll();
void ARM7opLDRptrmlr(), ARM7opLDRofrmlr(), ARM7opLDRprrmlr();
void ARM7opLDRptrplr(), ARM7opLDRofrplr(), ARM7opLDRprrplr();
void ARM7opLDRptrmar(), ARM7opLDRofrmar(), ARM7opLDRprrmar();
void ARM7opLDRptrpar(), ARM7opLDRofrpar(), ARM7opLDRprrpar();
void ARM7opLDRptrmrr(), ARM7opLDRofrmrr(), ARM7opLDRprrmrr();
void ARM7opLDRptrprr(), ARM7opLDRofrprr(), ARM7opLDRprrprr();

void ARM7opSTRptrmll(), ARM7opSTRofrmll(), ARM7opSTRprrmll();
void ARM7opSTRptrpll(), ARM7opSTRofrpll(), ARM7opSTRprrpll();
void ARM7opSTRptrmlr(), ARM7opSTRofrmlr(), ARM7opSTRprrmlr();
void ARM7opSTRptrplr(), ARM7opSTRofrplr(), ARM7opSTRprrplr();
void ARM7opSTRptrmar(), ARM7opSTRofrmar(), ARM7opSTRprrmar();
void ARM7opSTRptrpar(), ARM7opSTRofrpar(), ARM7opSTRprrpar();
void ARM7opSTRptrmrr(), ARM7opSTRofrmrr(), ARM7opSTRprrmrr();
void ARM7opSTRptrprr(), ARM7opSTRofrprr(), ARM7opSTRprrprr();

void ARM7opLDRBptrmll(), ARM7opLDRBofrmll(), ARM7opLDRBprrmll();
void ARM7opLDRBptrpll(), ARM7opLDRBofrpll(), ARM7opLDRBprrpll();
void ARM7opLDRBptrmlr(), ARM7opLDRBofrmlr(), ARM7opLDRBprrmlr();
void ARM7opLDRBptrplr(), ARM7opLDRBofrplr(), ARM7opLDRBprrplr();
void ARM7opLDRBptrmar(), ARM7opLDRBofrmar(), ARM7opLDRBprrmar();
void ARM7opLDRBptrpar(), ARM7opLDRBofrpar(), ARM7opLDRBprrpar();
void ARM7opLDRBptrmrr(), ARM7opLDRBofrmrr(), ARM7opLDRBprrmrr();
void ARM7opLDRBptrprr(), ARM7opLDRBofrprr(), ARM7opLDRBprrprr();

void ARM7opSTRBptrmll(), ARM7opSTRBofrmll(), ARM7opSTRBprrmll();
void ARM7opSTRBptrpll(), ARM7opSTRBofrpll(), ARM7opSTRBprrpll();
void ARM7opSTRBptrmlr(), ARM7opSTRBofrmlr(), ARM7opSTRBprrmlr();
void ARM7opSTRBptrplr(), ARM7opSTRBofrplr(), ARM7opSTRBprrplr();
void ARM7opSTRBptrmar(), ARM7opSTRBofrmar(), ARM7opSTRBprrmar();
void ARM7opSTRBptrpar(), ARM7opSTRBofrpar(), ARM7opSTRBprrpar();
void ARM7opSTRBptrmrr(), ARM7opSTRBofrmrr(), ARM7opSTRBprrmrr();
void ARM7opSTRBptrprr(), ARM7opSTRBofrprr(), ARM7opSTRBprrprr();

void ARM7opLDRTptrmll(), ARM7opLDRTofrmll(), ARM7opLDRTprrmll();
void ARM7opLDRTptrpll(), ARM7opLDRTofrpll(), ARM7opLDRTprrpll();
void ARM7opLDRTptrmlr(), ARM7opLDRTofrmlr(), ARM7opLDRTprrmlr();
void ARM7opLDRTptrplr(), ARM7opLDRTofrplr(), ARM7opLDRTprrplr();
void ARM7opLDRTptrmar(), ARM7opLDRTofrmar(), ARM7opLDRTprrmar();
void ARM7opLDRTptrpar(), ARM7opLDRTofrpar(), ARM7opLDRTprrpar();
void ARM7opLDRTptrmrr(), ARM7opLDRTofrmrr(), ARM7opLDRTprrmrr();
void ARM7opLDRTptrprr(), ARM7opLDRTofrprr(), ARM7opLDRTprrprr();

void ARM7opSTRTptrmll(), ARM7opSTRTofrmll(), ARM7opSTRTprrmll();
void ARM7opSTRTptrpll(), ARM7opSTRTofrpll(), ARM7opSTRTprrpll();
void ARM7opSTRTptrmlr(), ARM7opSTRTofrmlr(), ARM7opSTRTprrmlr();
void ARM7opSTRTptrplr(), ARM7opSTRTofrplr(), ARM7opSTRTprrplr();
void ARM7opSTRTptrmar(), ARM7opSTRTofrmar(), ARM7opSTRTprrmar();
void ARM7opSTRTptrpar(), ARM7opSTRTofrpar(), ARM7opSTRTprrpar();
void ARM7opSTRTptrmrr(), ARM7opSTRTofrmrr(), ARM7opSTRTprrmrr();
void ARM7opSTRTptrprr(), ARM7opSTRTofrprr(), ARM7opSTRTprrprr();

void ARM7opLDRBTptrmll(), ARM7opLDRBTofrmll(), ARM7opLDRBTprrmll();
void ARM7opLDRBTptrpll(), ARM7opLDRBTofrpll(), ARM7opLDRBTprrpll();
void ARM7opLDRBTptrmlr(), ARM7opLDRBTofrmlr(), ARM7opLDRBTprrmlr();
void ARM7opLDRBTptrplr(), ARM7opLDRBTofrplr(), ARM7opLDRBTprrplr();
void ARM7opLDRBTptrmar(), ARM7opLDRBTofrmar(), ARM7opLDRBTprrmar();
void ARM7opLDRBTptrpar(), ARM7opLDRBTofrpar(), ARM7opLDRBTprrpar();
void ARM7opLDRBTptrmrr(), ARM7opLDRBTofrmrr(), ARM7opLDRBTprrmrr();
void ARM7opLDRBTptrprr(), ARM7opLDRBTofrprr(), ARM7opLDRBTprrprr();

void ARM7opSTRBTptrmll(), ARM7opSTRBTofrmll(), ARM7opSTRBTprrmll();
void ARM7opSTRBTptrpll(), ARM7opSTRBTofrpll(), ARM7opSTRBTprrpll();
void ARM7opSTRBTptrmlr(), ARM7opSTRBTofrmlr(), ARM7opSTRBTprrmlr();
void ARM7opSTRBTptrplr(), ARM7opSTRBTofrplr(), ARM7opSTRBTprrplr();
void ARM7opSTRBTptrmar(), ARM7opSTRBTofrmar(), ARM7opSTRBTprrmar();
void ARM7opSTRBTptrpar(), ARM7opSTRBTofrpar(), ARM7opSTRBTprrpar();
void ARM7opSTRBTptrmrr(), ARM7opSTRBTofrmrr(), ARM7opSTRBTprrmrr();
void ARM7opSTRBTptrprr(), ARM7opSTRBTofrprr(), ARM7opSTRBTprrprr();

//---SWP-------------------------------------------------------------------

void ARM7opSWP(), ARM7opSWPB();

//---LoaD/STore Multiple; Increment/Decrement After/Before-----------------

void ARM7opLDMIA(), ARM7opLDMIB(), ARM7opLDMDA(), ARM7opLDMDB();
void ARM7opSTMIA(), ARM7opSTMIB(), ARM7opSTMDA(), ARM7opSTMDB();

void ARM7opLDMIAu(), ARM7opLDMIBu(), ARM7opLDMDAu(), ARM7opLDMDBu();
void ARM7opSTMIAu(), ARM7opSTMIBu(), ARM7opSTMDAu(), ARM7opSTMDBu();

void ARM7opLDMIAw(), ARM7opLDMIBw(), ARM7opLDMDAw(), ARM7opLDMDBw();
void ARM7opSTMIAw(), ARM7opSTMIBw(), ARM7opSTMDAw(), ARM7opSTMDBw();

void ARM7opLDMIAuw(), ARM7opLDMIBuw(), ARM7opLDMDAuw(), ARM7opLDMDBuw();
void ARM7opSTMIAuw(), ARM7opSTMIBuw(), ARM7opSTMDAuw(), ARM7opSTMDBuw();

//---SoftWare Interrupts---------------------------------------------------

void ARM7opSWI();

//---Thumb ops-------------------------------------------------------------

void Thumb7opUND();

void Thumb7opLSLimm(), Thumb7opLSRimm(), Thumb7opASRimm();
void Thumb7opADDreg(), Thumb7opSUBreg();
void Thumb7opADDimm3(), Thumb7opSUBimm3();

void Thumb7opMOVimm8r0(), Thumb7opMOVimm8r1();
void Thumb7opMOVimm8r2(), Thumb7opMOVimm8r3();
void Thumb7opMOVimm8r4(), Thumb7opMOVimm8r5();
void Thumb7opMOVimm8r6(), Thumb7opMOVimm8r7();
void Thumb7opCMPimm8r0(), Thumb7opCMPimm8r1();
void Thumb7opCMPimm8r2(), Thumb7opCMPimm8r3();
void Thumb7opCMPimm8r4(), Thumb7opCMPimm8r5();
void Thumb7opCMPimm8r6(), Thumb7opCMPimm8r7();
void Thumb7opADDimm8r0(), Thumb7opADDimm8r1();
void Thumb7opADDimm8r2(), Thumb7opADDimm8r3();
void Thumb7opADDimm8r4(), Thumb7opADDimm8r5();
void Thumb7opADDimm8r6(), Thumb7opADDimm8r7();
void Thumb7opSUBimm8r0(), Thumb7opSUBimm8r1();
void Thumb7opSUBimm8r2(), Thumb7opSUBimm8r3();
void Thumb7opSUBimm8r4(), Thumb7opSUBimm8r5();
void Thumb7opSUBimm8r6(), Thumb7opSUBimm8r7();

void Thumb7opDPg1(), Thumb7opDPg2(), Thumb7opDPg3(), Thumb7opDPg4();
void Thumb7opADDH(), Thumb7opMOVH(), Thumb7opCMPH(), Thumb7opBXreg();

void Thumb7opLDRPCr0(), Thumb7opLDRPCr1();
void Thumb7opLDRPCr2(), Thumb7opLDRPCr3();
void Thumb7opLDRPCr4(), Thumb7opLDRPCr5();
void Thumb7opLDRPCr6(), Thumb7opLDRPCr7();

void Thumb7opSTRSPr0(), Thumb7opSTRSPr1();
void Thumb7opSTRSPr2(), Thumb7opSTRSPr3();
void Thumb7opSTRSPr4(), Thumb7opSTRSPr5();
void Thumb7opSTRSPr6(), Thumb7opSTRSPr7();
void Thumb7opLDRSPr0(), Thumb7opLDRSPr1();
void Thumb7opLDRSPr2(), Thumb7opLDRSPr3();
void Thumb7opLDRSPr4(), Thumb7opLDRSPr5();
void Thumb7opLDRSPr6(), Thumb7opLDRSPr7();

void Thumb7opSTRreg(),  Thumb7opLDRreg();
void Thumb7opSTRHreg(), Thumb7opLDRHreg();
void Thumb7opSTRBreg(), Thumb7opLDRBreg();

void Thumb7opLDRSBreg(), Thumb7opLDRSHreg();

void Thumb7opSTRimm5(),  Thumb7opLDRimm5();
void Thumb7opSTRBimm5(), Thumb7opLDRBimm5();
void Thumb7opSTRHimm5(), Thumb7opLDRHimm5();

void Thumb7opADDPCr0(), Thumb7opADDPCr1();
void Thumb7opADDPCr2(), Thumb7opADDPCr3();
void Thumb7opADDPCr4(), Thumb7opADDPCr5();
void Thumb7opADDPCr6(), Thumb7opADDPCr7();
void Thumb7opADDSPr0(), Thumb7opADDSPr1();
void Thumb7opADDSPr2(), Thumb7opADDSPr3();
void Thumb7opADDSPr4(), Thumb7opADDSPr5();
void Thumb7opADDSPr6(), Thumb7opADDSPr7();
void Thumb7opADDSPimm7();

void Thumb7opPUSH(), Thumb7opPUSHlr();
void Thumb7opPOP(),  Thumb7opPOPpc();

void Thumb7opBKPT(), Thumb7opSWI();

void Thumb7opLDMIAr0(), Thumb7opLDMIAr1();
void Thumb7opLDMIAr2(), Thumb7opLDMIAr3();
void Thumb7opLDMIAr4(), Thumb7opLDMIAr5();
void Thumb7opLDMIAr6(), Thumb7opLDMIAr7();
void Thumb7opSTMIAr0(), Thumb7opSTMIAr1();
void Thumb7opSTMIAr2(), Thumb7opSTMIAr3();
void Thumb7opSTMIAr4(), Thumb7opSTMIAr5();
void Thumb7opSTMIAr6(), Thumb7opSTMIAr7();

void Thumb7opBEQ(), Thumb7opBNE();
void Thumb7opBCS(), Thumb7opBCC();
void Thumb7opBMI(), Thumb7opBPL();
void Thumb7opBVS(), Thumb7opBVC();
void Thumb7opBHI(), Thumb7opBLS();
void Thumb7opBGE(), Thumb7opBLT();
void Thumb7opBGT(), Thumb7opBLE();

void Thumb7opB(), Thumb7opBLsetup(), Thumb7opBLoff();

//---And so it begins------------------------------------------------------

typedef struct {
    fptr op;
    u32 cyc;
    u32 pad1,pad2;
} ARM7OPDESC;

//---Opcode function pointer table. Indexed by 27-20|7-4-------------------

const ARM7OPDESC arm7ops[]={
    /*0x00*/
    {ARM7opANDlli      ,2},
    {ARM7opANDllr      ,2},
    {ARM7opANDlri      ,2},
    {ARM7opANDlrr      ,2},
    {ARM7opANDari      ,2},
    {ARM7opANDarr      ,2},
    {ARM7opANDrri      ,2},
    {ARM7opANDrrr      ,2},
    {ARM7opANDlli      ,2},
    {ARM7opMUL         ,3},
    {ARM7opANDlri      ,2},
    {ARM7opSTRHptrm    ,2},
    {ARM7opANDari      ,2},
    {ARM7opUNP         ,1},  // Defined in 5E!
    {ARM7opANDrri      ,2},
    {ARM7opUNP         ,1},  // Defined in 5E!

    /*0x01*/
    {ARM7opANDSlli     ,2},
    {ARM7opANDSllr     ,2},
    {ARM7opANDSlri     ,2},
    {ARM7opANDSlrr     ,2},
    {ARM7opANDSari     ,2},
    {ARM7opANDSarr     ,2},
    {ARM7opANDSrri     ,2},
    {ARM7opANDSrrr     ,2},
    {ARM7opANDSlli     ,2},
    {ARM7opMULS        ,3},
    {ARM7opANDSlri     ,2},
    {ARM7opLDRHptrm    ,3},
    {ARM7opANDSari     ,2},
    {ARM7opLDRSBptrm   ,3},
    {ARM7opANDSrri     ,2},
    {ARM7opLDRSHptrm   ,3},

    /*0x02*/
    {ARM7opEORlli      ,2},
    {ARM7opEORllr      ,2},
    {ARM7opEORlri      ,2},
    {ARM7opEORlrr      ,2},
    {ARM7opEORari      ,2},
    {ARM7opEORarr      ,2},
    {ARM7opEORrri      ,2},
    {ARM7opEORrrr      ,2},
    {ARM7opEORlli      ,2},
    {ARM7opMLA         ,4},
    {ARM7opEORlri      ,2},
    {ARM7opSTRHptrm    ,2},  // Post-indexed with W bit
    {ARM7opEORari      ,2},
    {ARM7opUNP         ,1},  // Post-indexed with W bit
    {ARM7opEORrri      ,2},
    {ARM7opUNL         ,1},  // Post-indexed with W bit [reindexed]

    /*0x03*/
    {ARM7opEORSlli     ,2},
    {ARM7opEORSllr     ,2},
    {ARM7opEORSlri     ,2},
    {ARM7opEORSlrr     ,2},
    {ARM7opEORSari     ,2},
    {ARM7opEORSarr     ,2},
    {ARM7opEORSrri     ,2},
    {ARM7opEORSrrr     ,2},
    {ARM7opEORSlli     ,2},
    {ARM7opMLAS        ,4},
    {ARM7opEORSlri     ,2},
    {ARM7opLDRHptrm    ,3},  // Post-indexed with W bit
    {ARM7opEORSari     ,2},
    {ARM7opLDRSBptrm   ,3},  // Post-indexed with W bit
    {ARM7opEORSrri     ,2},
    {ARM7opLDRSHptrm   ,3},  // Post-indexed with W bit

    /*0x04*/
    {ARM7opSUBlli      ,2},
    {ARM7opSUBllr      ,2},
    {ARM7opSUBlri      ,2},
    {ARM7opSUBlrr      ,2},
    {ARM7opSUBari      ,2},
    {ARM7opSUBarr      ,2},
    {ARM7opSUBrri      ,2},
    {ARM7opSUBrrr      ,2},
    {ARM7opSUBlli      ,2},
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opSUBlri      ,2},
    {ARM7opSTRHptim    ,2},
    {ARM7opSUBari      ,2},
    {ARM7opUNP         ,1},  // Defined in 5E!
    {ARM7opSUBrri      ,2},
    {ARM7opUNP         ,1},  // Defined in 5E!

    /*0x05*/
    {ARM7opSUBSlli     ,2},
    {ARM7opSUBSllr     ,2},
    {ARM7opSUBSlri     ,2},
    {ARM7opSUBSlrr     ,2},
    {ARM7opSUBSari     ,2},
    {ARM7opSUBSarr     ,2},
    {ARM7opSUBSrri     ,2},
    {ARM7opSUBSrrr     ,2},
    {ARM7opSUBSlli     ,2},
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opSUBSlri     ,2},
    {ARM7opLDRHptim    ,3},
    {ARM7opSUBSari     ,2},
    {ARM7opLDRSBptim   ,3},
    {ARM7opSUBSrri     ,2},
    {ARM7opLDRSHptim   ,3},

    /*0x06*/
    {ARM7opRSBlli      ,2},
    {ARM7opRSBllr      ,2},
    {ARM7opRSBlri      ,2},
    {ARM7opRSBlrr      ,2},
    {ARM7opRSBari      ,2},
    {ARM7opRSBarr      ,2},
    {ARM7opRSBrri      ,2},
    {ARM7opRSBrrr      ,2},
    {ARM7opRSBlli      ,2},
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opRSBlri      ,2},
    {ARM7opSTRHptim    ,2},  // Post-indexed with W bit
    {ARM7opRSBari      ,2},
    {ARM7opUNP         ,1},  // Post-indexed with W bit
    {ARM7opRSBrri      ,2},
    {ARM7opUNP         ,1},  // Post-indexed with W bit

    /*0x07*/
    {ARM7opRSBSlli     ,2},
    {ARM7opRSBSllr     ,2},
    {ARM7opRSBSlri     ,2},
    {ARM7opRSBSlrr     ,2},
    {ARM7opRSBSari     ,2},
    {ARM7opRSBSarr     ,2},
    {ARM7opRSBSrri     ,2},
    {ARM7opRSBSrrr     ,2},
    {ARM7opRSBSlli     ,2},
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opRSBSlri     ,2},
    {ARM7opLDRHptim    ,3},  // Post-indexed with W bit
    {ARM7opRSBSari     ,2},
    {ARM7opLDRSBptim   ,3},  // Post-indexed with W bit
    {ARM7opRSBSrri     ,2},
    {ARM7opLDRSHptim   ,3},  // Post-indexed with W bit

    /*0x08*/
    {ARM7opADDlli      ,2},
    {ARM7opADDllr      ,2},
    {ARM7opADDlri      ,2},
    {ARM7opADDlrr      ,2},
    {ARM7opADDari      ,2},
    {ARM7opADDarr      ,2},
    {ARM7opADDrri      ,2},
    {ARM7opADDrrr      ,2},
    {ARM7opADDlli      ,2},
    {ARM7opUMULL       ,4},
    {ARM7opADDlri      ,2},
    {ARM7opSTRHptrp    ,2},
    {ARM7opADDari      ,2},
    {ARM7opUNP         ,1},  // Defined in 5E!
    {ARM7opADDrri      ,2},
    {ARM7opUNP         ,1},  // Defined in 5E!

    /*0x09*/
    {ARM7opADDSlli     ,2},
    {ARM7opADDSllr     ,2},
    {ARM7opADDSlri     ,2},
    {ARM7opADDSlrr     ,2},
    {ARM7opADDSari     ,2},
    {ARM7opADDSarr     ,2},
    {ARM7opADDSrri     ,2},
    {ARM7opADDSrrr     ,2},
    {ARM7opADDSlli     ,2},
    {ARM7opUMULLS      ,4},
    {ARM7opADDSlri     ,2},
    {ARM7opLDRHptrp    ,3},
    {ARM7opADDSari     ,2},
    {ARM7opLDRSBptrp   ,3},
    {ARM7opADDSrri     ,2},
    {ARM7opLDRSHptrp   ,3},

    /*0x0A*/
    {ARM7opADClli      ,2},
    {ARM7opADCllr      ,2},
    {ARM7opADClri      ,2},
    {ARM7opADClrr      ,2},
    {ARM7opADCari      ,2},
    {ARM7opADCarr      ,2},
    {ARM7opADCrri      ,2},
    {ARM7opADCrrr      ,2},
    {ARM7opADClli      ,2},
    {ARM7opUMLAL       ,5},
    {ARM7opADClri      ,2},
    {ARM7opSTRHptrp    ,2},  // Post-indexed with W bit
    {ARM7opADCari      ,2},
    {ARM7opUNP         ,1},  // Post-indexed with W bit
    {ARM7opADCrri      ,2},
    {ARM7opUNP         ,1},  // Post-indexed with W bit

    /*0x0B*/
    {ARM7opADCSlli     ,2},
    {ARM7opADCSllr     ,2},
    {ARM7opADCSlri     ,2},
    {ARM7opADCSlrr     ,2},
    {ARM7opADCSari     ,2},
    {ARM7opADCSarr     ,2},
    {ARM7opADCSrri     ,2},
    {ARM7opADCSrrr     ,2},
    {ARM7opADCSlli     ,2},
    {ARM7opUMLALS      ,5},
    {ARM7opADCSlri     ,2},
    {ARM7opLDRHptrp    ,3},  // Post-indexed with W bit
    {ARM7opADCSari     ,2},
    {ARM7opLDRSBptrp   ,3},  // Post-indexed with W bit
    {ARM7opADCSrri     ,2},
    {ARM7opLDRSHptrp   ,3},  // Post-indexed with W bit

    /*0x0C*/
    {ARM7opSBClli      ,2},
    {ARM7opSBCllr      ,2},
    {ARM7opSBClri      ,2},
    {ARM7opSBClrr      ,2},
    {ARM7opSBCari      ,2},
    {ARM7opSBCarr      ,2},
    {ARM7opSBCrri      ,2},
    {ARM7opSBCrrr      ,2},
    {ARM7opSBClli      ,2},
    {ARM7opSMULL       ,4},
    {ARM7opSBClri      ,2},
    {ARM7opSTRHptip    ,2},
    {ARM7opSBCari      ,2},
    {ARM7opUNP         ,1},  // Defined in 5E!
    {ARM7opSBCrri      ,2},
    {ARM7opUNP         ,1},  // Defined in 5E!

    /*0x0D*/
    {ARM7opSBCSlli     ,2},
    {ARM7opSBCSllr     ,2},
    {ARM7opSBCSlri     ,2},
    {ARM7opSBCSlrr     ,2},
    {ARM7opSBCSari     ,2},
    {ARM7opSBCSarr     ,2},
    {ARM7opSBCSrri     ,2},
    {ARM7opSBCSrrr     ,2},
    {ARM7opSBCSlli     ,2},
    {ARM7opSMULLS      ,4},
    {ARM7opSBCSlri     ,2},
    {ARM7opLDRHptip    ,3},
    {ARM7opSBCSari     ,2},
    {ARM7opLDRSBptip   ,3},
    {ARM7opSBCSrri     ,2},
    {ARM7opLDRSHptip   ,3},

    /*0x0E*/
    {ARM7opRSClli      ,2},
    {ARM7opRSCllr      ,2},
    {ARM7opRSClri      ,2},
    {ARM7opRSClrr      ,2},
    {ARM7opRSCari      ,2},
    {ARM7opRSCarr      ,2},
    {ARM7opRSCrri      ,2},
    {ARM7opRSCrrr      ,2},
    {ARM7opRSClli      ,2},
    {ARM7opSMLAL       ,5},
    {ARM7opRSClri      ,2},
    {ARM7opSTRHptip    ,2},  // Post-indexed with W bit
    {ARM7opRSCari      ,2},
    {ARM7opUNP         ,1},  // Post-indexed with W bit
    {ARM7opRSCrri      ,2},
    {ARM7opUNP         ,1},  // Post-indexed with W bit

    /*0x0F*/
    {ARM7opRSCSlli     ,2},
    {ARM7opRSCSllr     ,2},
    {ARM7opRSCSlri     ,2},
    {ARM7opRSCSlrr     ,2},
    {ARM7opRSCSari     ,2},
    {ARM7opRSCSarr     ,2},
    {ARM7opRSCSrri     ,2},
    {ARM7opRSCSrrr     ,2},
    {ARM7opRSCSlli     ,2},
    {ARM7opSMLALS      ,5},
    {ARM7opRSCSlri     ,2},
    {ARM7opLDRHptip    ,3},  // Post-indexed with W bit
    {ARM7opRSCSari     ,2},
    {ARM7opLDRSBptip   ,3},  // Post-indexed with W bit
    {ARM7opRSCSrri     ,2},
    {ARM7opLDRSHptip   ,3},  // Post-indexed with W bit

    /*0x10*/
    {ARM7opMRSrc       ,1},
    {ARM7opUNP         ,1},  // MRS with non-zero 7-4
    {ARM7opUNP         ,1},  // MRS with non-zero 7-4
    {ARM7opUNP         ,1},  // MRS with non-zero 7-4
    {ARM7opUNP         ,1},  // MRS with non-zero 7-4
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opUNP         ,1},  // MRS with non-zero 7-4
    {ARM7opUNP         ,1},  // MRS with non-zero 7-4
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opSWP         ,4},
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opSTRHofrm    ,2},
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opUNP         ,1},  // Defined in ARM5E!

    /*0x11*/
    {ARM7opTSTSlli     ,2},
    {ARM7opTSTSllr     ,2},
    {ARM7opTSTSlri     ,2},
    {ARM7opTSTSlrr     ,2},
    {ARM7opTSTSari     ,2},
    {ARM7opTSTSarr     ,2},
    {ARM7opTSTSrri     ,2},
    {ARM7opTSTSrrr     ,2},
    {ARM7opTSTSlli     ,2},
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opTSTSlri     ,2},
    {ARM7opLDRHofrm    ,3},
    {ARM7opTSTSari     ,2},
    {ARM7opLDRSBofrm   ,3},
    {ARM7opTSTSrri     ,2},
    {ARM7opLDRSHofrm   ,3},

    /*0x12*/
    {ARM7opMSRrc       ,1},
    {ARM7opBX          ,3},
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opBLXr        ,3},
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opSTRHprrm    ,2},
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opUNP         ,1},  // Defined in ARM5E!

    /*0x13*/
    {ARM7opTEQSlli     ,2},
    {ARM7opTEQSllr     ,2},
    {ARM7opTEQSlri     ,2},
    {ARM7opTEQSlrr     ,2},
    {ARM7opTEQSari     ,2},
    {ARM7opTEQSarr     ,2},
    {ARM7opTEQSrri     ,2},
    {ARM7opTEQSrrr     ,2},
    {ARM7opTEQSlli     ,2},
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opTEQSlri     ,2},
    {ARM7opLDRHprrm    ,3},
    {ARM7opTEQSari     ,2},
    {ARM7opLDRSBprrm   ,3},
    {ARM7opTEQSrri     ,2},
    {ARM7opLDRSHprrm   ,3},

    /*0x14*/
    {ARM7opMRSrs       ,1},
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opSWPB        ,4},
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opSTRHofim    ,2},
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opUNP         ,1},  // Defined in ARM5E!

    /*0x15*/
    {ARM7opCMPSlli     ,2},
    {ARM7opCMPSllr     ,2},
    {ARM7opCMPSlri     ,2},
    {ARM7opCMPSlrr     ,2},
    {ARM7opCMPSari     ,2},
    {ARM7opCMPSarr     ,2},
    {ARM7opCMPSrri     ,2},
    {ARM7opCMPSrrr     ,2},
    {ARM7opCMPSlli     ,2},
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opCMPSlri     ,2},
    {ARM7opLDRHofim    ,3},
    {ARM7opCMPSari     ,2},
    {ARM7opLDRSBofim   ,3},
    {ARM7opCMPSrri     ,2},
    {ARM7opLDRSHofim   ,3},

    /*0x16*/
    {ARM7opMSRrs       ,1},
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opSTRHprim    ,2},
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opUNP         ,1},  // Defined in ARM5E!
    {ARM7opUNP         ,1},  // Defined in ARM5E!

    /*0x17*/
    {ARM7opCMNSlli     ,2},
    {ARM7opCMNSllr     ,2},
    {ARM7opCMNSlri     ,2},
    {ARM7opCMNSlrr     ,2},
    {ARM7opCMNSari     ,2},
    {ARM7opCMNSarr     ,2},
    {ARM7opCMNSrri     ,2},
    {ARM7opCMNSrrr     ,2},
    {ARM7opCMNSlli     ,2},
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opCMNSlri     ,2},
    {ARM7opLDRHprim    ,3},
    {ARM7opCMNSari     ,2},
    {ARM7opLDRSBprim   ,3},
    {ARM7opCMNSrri     ,2},
    {ARM7opLDRSHprim   ,3},

    /*0x18*/
    {ARM7opORRlli      ,2},
    {ARM7opORRllr      ,2},
    {ARM7opORRlri      ,2},
    {ARM7opORRlrr      ,2},
    {ARM7opORRari      ,2},
    {ARM7opORRarr      ,2},
    {ARM7opORRrri      ,2},
    {ARM7opORRrrr      ,2},
    {ARM7opORRlli      ,2},
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opORRlri      ,2},
    {ARM7opSTRHofrp    ,2},
    {ARM7opORRari      ,2},
    {ARM7opUNP         ,1},  // Defined in 5E!
    {ARM7opORRrri      ,2},
    {ARM7opUNP         ,1},  // Defined in 5E!

    /*0x19*/
    {ARM7opORRSlli     ,2},
    {ARM7opORRSllr     ,2},
    {ARM7opORRSlri     ,2},
    {ARM7opORRSlrr     ,2},
    {ARM7opORRSari     ,2},
    {ARM7opORRSarr     ,2},
    {ARM7opORRSrri     ,2},
    {ARM7opORRSrrr     ,2},
    {ARM7opORRSlli     ,2},
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opORRSlri     ,2},
    {ARM7opLDRHofrp    ,3},
    {ARM7opORRSari     ,2},
    {ARM7opLDRSBofrp   ,3},
    {ARM7opORRSrri     ,2},
    {ARM7opLDRSHofrp   ,3},

    /*0x1A*/
    {ARM7opMOVlli      ,2},
    {ARM7opMOVllr      ,2},
    {ARM7opMOVlri      ,2},
    {ARM7opMOVlrr      ,2},
    {ARM7opMOVari      ,2},
    {ARM7opMOVarr      ,2},
    {ARM7opMOVrri      ,2},
    {ARM7opMOVrrr      ,2},
    {ARM7opMOVlli      ,2},
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opMOVlri      ,2},
    {ARM7opSTRHprrp    ,2},
    {ARM7opMOVari      ,2},
    {ARM7opUNP         ,1},  // Defined in 5E!
    {ARM7opMOVrri      ,2},
    {ARM7opUNP         ,1},  // Defined in 5E!

    /*0x1B*/
    {ARM7opMOVSlli     ,2},
    {ARM7opMOVSllr     ,2},
    {ARM7opMOVSlri     ,2},
    {ARM7opMOVSlrr     ,2},
    {ARM7opMOVSari     ,2},
    {ARM7opMOVSarr     ,2},
    {ARM7opMOVSrri     ,2},
    {ARM7opMOVSrrr     ,2},
    {ARM7opMOVSlli     ,2},
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opMOVSlri     ,2},
    {ARM7opLDRHprrp    ,3},
    {ARM7opMOVSari     ,2},
    {ARM7opLDRSBprrp   ,3},
    {ARM7opMOVSrri     ,2},
    {ARM7opLDRSHprrp   ,3},

    /*0x1C*/
    {ARM7opBIClli      ,2},
    {ARM7opBICllr      ,2},
    {ARM7opBIClri      ,2},
    {ARM7opBIClrr      ,2},
    {ARM7opBICari      ,2},
    {ARM7opBICarr      ,2},
    {ARM7opBICrri      ,2},
    {ARM7opBICrrr      ,2},
    {ARM7opBIClli      ,2},
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opBIClri      ,2},
    {ARM7opSTRHofip    ,2},
    {ARM7opBICari      ,2},
    {ARM7opUNP         ,1},  // Defined in 5E!
    {ARM7opBICrri      ,2},
    {ARM7opUNP         ,1},  // Defined in 5E!

    /*0x1D*/
    {ARM7opBICSlli     ,2},
    {ARM7opBICSllr     ,2},
    {ARM7opBICSlri     ,2},
    {ARM7opBICSlrr     ,2},
    {ARM7opBICSari     ,2},
    {ARM7opBICSarr     ,2},
    {ARM7opBICSrri     ,2},
    {ARM7opBICSrrr     ,2},
    {ARM7opBICSlli     ,2},
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opBICSlri     ,2},
    {ARM7opLDRHofip    ,3},
    {ARM7opBICSari     ,2},
    {ARM7opLDRSBofip   ,3},
    {ARM7opBICSrri     ,2},
    {ARM7opLDRSHofip   ,3},

    /*0x1E*/
    {ARM7opMVNlli      ,2},
    {ARM7opMVNllr      ,2},
    {ARM7opMVNlri      ,2},
    {ARM7opMVNlrr      ,2},
    {ARM7opMVNari      ,2},
    {ARM7opMVNarr      ,2},
    {ARM7opMVNrri      ,2},
    {ARM7opMVNrrr      ,2},
    {ARM7opMVNlli      ,2},
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opMVNlri      ,2},
    {ARM7opSTRHprip    ,2},
    {ARM7opMVNari      ,2},
    {ARM7opUNP         ,1},  // Defined in 5E!
    {ARM7opMVNrri      ,2},
    {ARM7opUNP         ,1},  // Defined in 5E!

    /*0x1F*/
    {ARM7opMVNSlli     ,2},
    {ARM7opMVNSllr     ,2},
    {ARM7opMVNSlri     ,2},
    {ARM7opMVNSlrr     ,2},
    {ARM7opMVNSari     ,2},
    {ARM7opMVNSarr     ,2},
    {ARM7opMVNSrri     ,2},
    {ARM7opMVNSrrr     ,2},
    {ARM7opMVNSlli     ,2},
    {ARM7opUNP         ,1},  // Unallocated
    {ARM7opMVNSlri     ,2},
    {ARM7opLDRHprip    ,3},
    {ARM7opMVNSari     ,2},
    {ARM7opLDRSBprip   ,3},
    {ARM7opMVNSrri     ,2},
    {ARM7opLDRSHprip   ,3},

    /*0x20*/
    {ARM7opANDimm      ,1},
    {ARM7opANDimm      ,1},
    {ARM7opANDimm      ,1},
    {ARM7opANDimm      ,1},
    {ARM7opANDimm      ,1},
    {ARM7opANDimm      ,1},
    {ARM7opANDimm      ,1},
    {ARM7opANDimm      ,1},
    {ARM7opANDimm      ,1},
    {ARM7opANDimm      ,1},
    {ARM7opANDimm      ,1},
    {ARM7opANDimm      ,1},
    {ARM7opANDimm      ,1},
    {ARM7opANDimm      ,1},
    {ARM7opANDimm      ,1},
    {ARM7opANDimm      ,1},

    /*0x21*/
    {ARM7opANDSimm     ,1},
    {ARM7opANDSimm     ,1},
    {ARM7opANDSimm     ,1},
    {ARM7opANDSimm     ,1},
    {ARM7opANDSimm     ,1},
    {ARM7opANDSimm     ,1},
    {ARM7opANDSimm     ,1},
    {ARM7opANDSimm     ,1},
    {ARM7opANDSimm     ,1},
    {ARM7opANDSimm     ,1},
    {ARM7opANDSimm     ,1},
    {ARM7opANDSimm     ,1},
    {ARM7opANDSimm     ,1},
    {ARM7opANDSimm     ,1},
    {ARM7opANDSimm     ,1},
    {ARM7opANDSimm     ,1},

    /*0x22*/
    {ARM7opEORimm      ,1},
    {ARM7opEORimm      ,1},
    {ARM7opEORimm      ,1},
    {ARM7opEORimm      ,1},
    {ARM7opEORimm      ,1},
    {ARM7opEORimm      ,1},
    {ARM7opEORimm      ,1},
    {ARM7opEORimm      ,1},
    {ARM7opEORimm      ,1},
    {ARM7opEORimm      ,1},
    {ARM7opEORimm      ,1},
    {ARM7opEORimm      ,1},
    {ARM7opEORimm      ,1},
    {ARM7opEORimm      ,1},
    {ARM7opEORimm      ,1},
    {ARM7opEORimm      ,1},

    /*0x23*/
    {ARM7opEORSimm     ,1},
    {ARM7opEORSimm     ,1},
    {ARM7opEORSimm     ,1},
    {ARM7opEORSimm     ,1},
    {ARM7opEORSimm     ,1},
    {ARM7opEORSimm     ,1},
    {ARM7opEORSimm     ,1},
    {ARM7opEORSimm     ,1},
    {ARM7opEORSimm     ,1},
    {ARM7opEORSimm     ,1},
    {ARM7opEORSimm     ,1},
    {ARM7opEORSimm     ,1},
    {ARM7opEORSimm     ,1},
    {ARM7opEORSimm     ,1},
    {ARM7opEORSimm     ,1},
    {ARM7opEORSimm     ,1},

    /*0x24*/
    {ARM7opSUBimm      ,1},
    {ARM7opSUBimm      ,1},
    {ARM7opSUBimm      ,1},
    {ARM7opSUBimm      ,1},
    {ARM7opSUBimm      ,1},
    {ARM7opSUBimm      ,1},
    {ARM7opSUBimm      ,1},
    {ARM7opSUBimm      ,1},
    {ARM7opSUBimm      ,1},
    {ARM7opSUBimm      ,1},
    {ARM7opSUBimm      ,1},
    {ARM7opSUBimm      ,1},
    {ARM7opSUBimm      ,1},
    {ARM7opSUBimm      ,1},
    {ARM7opSUBimm      ,1},
    {ARM7opSUBimm      ,1},

    /*0x25*/
    {ARM7opSUBSimm     ,1},
    {ARM7opSUBSimm     ,1},
    {ARM7opSUBSimm     ,1},
    {ARM7opSUBSimm     ,1},
    {ARM7opSUBSimm     ,1},
    {ARM7opSUBSimm     ,1},
    {ARM7opSUBSimm     ,1},
    {ARM7opSUBSimm     ,1},
    {ARM7opSUBSimm     ,1},
    {ARM7opSUBSimm     ,1},
    {ARM7opSUBSimm     ,1},
    {ARM7opSUBSimm     ,1},
    {ARM7opSUBSimm     ,1},
    {ARM7opSUBSimm     ,1},
    {ARM7opSUBSimm     ,1},
    {ARM7opSUBSimm     ,1},

    /*0x26*/
    {ARM7opRSBimm      ,1},
    {ARM7opRSBimm      ,1},
    {ARM7opRSBimm      ,1},
    {ARM7opRSBimm      ,1},
    {ARM7opRSBimm      ,1},
    {ARM7opRSBimm      ,1},
    {ARM7opRSBimm      ,1},
    {ARM7opRSBimm      ,1},
    {ARM7opRSBimm      ,1},
    {ARM7opRSBimm      ,1},
    {ARM7opRSBimm      ,1},
    {ARM7opRSBimm      ,1},
    {ARM7opRSBimm      ,1},
    {ARM7opRSBimm      ,1},
    {ARM7opRSBimm      ,1},
    {ARM7opRSBimm      ,1},

    /*0x27*/
    {ARM7opRSBSimm     ,1},
    {ARM7opRSBSimm     ,1},
    {ARM7opRSBSimm     ,1},
    {ARM7opRSBSimm     ,1},
    {ARM7opRSBSimm     ,1},
    {ARM7opRSBSimm     ,1},
    {ARM7opRSBSimm     ,1},
    {ARM7opRSBSimm     ,1},
    {ARM7opRSBSimm     ,1},
    {ARM7opRSBSimm     ,1},
    {ARM7opRSBSimm     ,1},
    {ARM7opRSBSimm     ,1},
    {ARM7opRSBSimm     ,1},
    {ARM7opRSBSimm     ,1},
    {ARM7opRSBSimm     ,1},
    {ARM7opRSBSimm     ,1},

    /*0x28*/
    {ARM7opADDimm      ,1},
    {ARM7opADDimm      ,1},
    {ARM7opADDimm      ,1},
    {ARM7opADDimm      ,1},
    {ARM7opADDimm      ,1},
    {ARM7opADDimm      ,1},
    {ARM7opADDimm      ,1},
    {ARM7opADDimm      ,1},
    {ARM7opADDimm      ,1},
    {ARM7opADDimm      ,1},
    {ARM7opADDimm      ,1},
    {ARM7opADDimm      ,1},
    {ARM7opADDimm      ,1},
    {ARM7opADDimm      ,1},
    {ARM7opADDimm      ,1},
    {ARM7opADDimm      ,1},

    /*0x29*/
    {ARM7opADDSimm     ,1},
    {ARM7opADDSimm     ,1},
    {ARM7opADDSimm     ,1},
    {ARM7opADDSimm     ,1},
    {ARM7opADDSimm     ,1},
    {ARM7opADDSimm     ,1},
    {ARM7opADDSimm     ,1},
    {ARM7opADDSimm     ,1},
    {ARM7opADDSimm     ,1},
    {ARM7opADDSimm     ,1},
    {ARM7opADDSimm     ,1},
    {ARM7opADDSimm     ,1},
    {ARM7opADDSimm     ,1},
    {ARM7opADDSimm     ,1},
    {ARM7opADDSimm     ,1},
    {ARM7opADDSimm     ,1},

    /*0x2A*/
    {ARM7opADCimm      ,1},
    {ARM7opADCimm      ,1},
    {ARM7opADCimm      ,1},
    {ARM7opADCimm      ,1},
    {ARM7opADCimm      ,1},
    {ARM7opADCimm      ,1},
    {ARM7opADCimm      ,1},
    {ARM7opADCimm      ,1},
    {ARM7opADCimm      ,1},
    {ARM7opADCimm      ,1},
    {ARM7opADCimm      ,1},
    {ARM7opADCimm      ,1},
    {ARM7opADCimm      ,1},
    {ARM7opADCimm      ,1},
    {ARM7opADCimm      ,1},
    {ARM7opADCimm      ,1},

    /*0x2B*/
    {ARM7opADCSimm     ,1},
    {ARM7opADCSimm     ,1},
    {ARM7opADCSimm     ,1},
    {ARM7opADCSimm     ,1},
    {ARM7opADCSimm     ,1},
    {ARM7opADCSimm     ,1},
    {ARM7opADCSimm     ,1},
    {ARM7opADCSimm     ,1},
    {ARM7opADCSimm     ,1},
    {ARM7opADCSimm     ,1},
    {ARM7opADCSimm     ,1},
    {ARM7opADCSimm     ,1},
    {ARM7opADCSimm     ,1},
    {ARM7opADCSimm     ,1},
    {ARM7opADCSimm     ,1},
    {ARM7opADCSimm     ,1},

    /*0x2C*/
    {ARM7opSBCimm      ,1},
    {ARM7opSBCimm      ,1},
    {ARM7opSBCimm      ,1},
    {ARM7opSBCimm      ,1},
    {ARM7opSBCimm      ,1},
    {ARM7opSBCimm      ,1},
    {ARM7opSBCimm      ,1},
    {ARM7opSBCimm      ,1},
    {ARM7opSBCimm      ,1},
    {ARM7opSBCimm      ,1},
    {ARM7opSBCimm      ,1},
    {ARM7opSBCimm      ,1},
    {ARM7opSBCimm      ,1},
    {ARM7opSBCimm      ,1},
    {ARM7opSBCimm      ,1},
    {ARM7opSBCimm      ,1},

    /*0x2D*/
    {ARM7opSBCSimm     ,1},
    {ARM7opSBCSimm     ,1},
    {ARM7opSBCSimm     ,1},
    {ARM7opSBCSimm     ,1},
    {ARM7opSBCSimm     ,1},
    {ARM7opSBCSimm     ,1},
    {ARM7opSBCSimm     ,1},
    {ARM7opSBCSimm     ,1},
    {ARM7opSBCSimm     ,1},
    {ARM7opSBCSimm     ,1},
    {ARM7opSBCSimm     ,1},
    {ARM7opSBCSimm     ,1},
    {ARM7opSBCSimm     ,1},
    {ARM7opSBCSimm     ,1},
    {ARM7opSBCSimm     ,1},
    {ARM7opSBCSimm     ,1},

    /*0x2E*/
    {ARM7opRSCimm      ,1},
    {ARM7opRSCimm      ,1},
    {ARM7opRSCimm      ,1},
    {ARM7opRSCimm      ,1},
    {ARM7opRSCimm      ,1},
    {ARM7opRSCimm      ,1},
    {ARM7opRSCimm      ,1},
    {ARM7opRSCimm      ,1},
    {ARM7opRSCimm      ,1},
    {ARM7opRSCimm      ,1},
    {ARM7opRSCimm      ,1},
    {ARM7opRSCimm      ,1},
    {ARM7opRSCimm      ,1},
    {ARM7opRSCimm      ,1},
    {ARM7opRSCimm      ,1},
    {ARM7opRSCimm      ,1},

    /*0x2F*/
    {ARM7opRSCSimm     ,1},
    {ARM7opRSCSimm     ,1},
    {ARM7opRSCSimm     ,1},
    {ARM7opRSCSimm     ,1},
    {ARM7opRSCSimm     ,1},
    {ARM7opRSCSimm     ,1},
    {ARM7opRSCSimm     ,1},
    {ARM7opRSCSimm     ,1},
    {ARM7opRSCSimm     ,1},
    {ARM7opRSCSimm     ,1},
    {ARM7opRSCSimm     ,1},
    {ARM7opRSCSimm     ,1},
    {ARM7opRSCSimm     ,1},
    {ARM7opRSCSimm     ,1},
    {ARM7opRSCSimm     ,1},
    {ARM7opRSCSimm     ,1},

    /*0x30*/
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},

    /*0x31*/
    {ARM7opTSTSimm     ,1},
    {ARM7opTSTSimm     ,1},
    {ARM7opTSTSimm     ,1},
    {ARM7opTSTSimm     ,1},
    {ARM7opTSTSimm     ,1},
    {ARM7opTSTSimm     ,1},
    {ARM7opTSTSimm     ,1},
    {ARM7opTSTSimm     ,1},
    {ARM7opTSTSimm     ,1},
    {ARM7opTSTSimm     ,1},
    {ARM7opTSTSimm     ,1},
    {ARM7opTSTSimm     ,1},
    {ARM7opTSTSimm     ,1},
    {ARM7opTSTSimm     ,1},
    {ARM7opTSTSimm     ,1},
    {ARM7opTSTSimm     ,1},

    /*0x32*/
    {ARM7opMSRic       ,1},
    {ARM7opMSRic       ,1},
    {ARM7opMSRic       ,1},
    {ARM7opMSRic       ,1},
    {ARM7opMSRic       ,1},
    {ARM7opMSRic       ,1},
    {ARM7opMSRic       ,1},
    {ARM7opMSRic       ,1},
    {ARM7opMSRic       ,1},
    {ARM7opMSRic       ,1},
    {ARM7opMSRic       ,1},
    {ARM7opMSRic       ,1},
    {ARM7opMSRic       ,1},
    {ARM7opMSRic       ,1},
    {ARM7opMSRic       ,1},
    {ARM7opMSRic       ,1},

    /*0x33*/
    {ARM7opTEQSimm     ,1},
    {ARM7opTEQSimm     ,1},
    {ARM7opTEQSimm     ,1},
    {ARM7opTEQSimm     ,1},
    {ARM7opTEQSimm     ,1},
    {ARM7opTEQSimm     ,1},
    {ARM7opTEQSimm     ,1},
    {ARM7opTEQSimm     ,1},
    {ARM7opTEQSimm     ,1},
    {ARM7opTEQSimm     ,1},
    {ARM7opTEQSimm     ,1},
    {ARM7opTEQSimm     ,1},
    {ARM7opTEQSimm     ,1},
    {ARM7opTEQSimm     ,1},
    {ARM7opTEQSimm     ,1},
    {ARM7opTEQSimm     ,1},

    /*0x34*/
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},
    {ARM7opUND         ,4},

    /*0x35*/
    {ARM7opCMPSimm     ,1},
    {ARM7opCMPSimm     ,1},
    {ARM7opCMPSimm     ,1},
    {ARM7opCMPSimm     ,1},
    {ARM7opCMPSimm     ,1},
    {ARM7opCMPSimm     ,1},
    {ARM7opCMPSimm     ,1},
    {ARM7opCMPSimm     ,1},
    {ARM7opCMPSimm     ,1},
    {ARM7opCMPSimm     ,1},
    {ARM7opCMPSimm     ,1},
    {ARM7opCMPSimm     ,1},
    {ARM7opCMPSimm     ,1},
    {ARM7opCMPSimm     ,1},
    {ARM7opCMPSimm     ,1},
    {ARM7opCMPSimm     ,1},

    /*0x36*/
    {ARM7opMSRis       ,1},
    {ARM7opMSRis       ,1},
    {ARM7opMSRis       ,1},
    {ARM7opMSRis       ,1},
    {ARM7opMSRis       ,1},
    {ARM7opMSRis       ,1},
    {ARM7opMSRis       ,1},
    {ARM7opMSRis       ,1},
    {ARM7opMSRis       ,1},
    {ARM7opMSRis       ,1},
    {ARM7opMSRis       ,1},
    {ARM7opMSRis       ,1},
    {ARM7opMSRis       ,1},
    {ARM7opMSRis       ,1},
    {ARM7opMSRis       ,1},
    {ARM7opMSRis       ,1},

    /*0x37*/
    {ARM7opCMNSimm     ,1},
    {ARM7opCMNSimm     ,1},
    {ARM7opCMNSimm     ,1},
    {ARM7opCMNSimm     ,1},
    {ARM7opCMNSimm     ,1},
    {ARM7opCMNSimm     ,1},
    {ARM7opCMNSimm     ,1},
    {ARM7opCMNSimm     ,1},
    {ARM7opCMNSimm     ,1},
    {ARM7opCMNSimm     ,1},
    {ARM7opCMNSimm     ,1},
    {ARM7opCMNSimm     ,1},
    {ARM7opCMNSimm     ,1},
    {ARM7opCMNSimm     ,1},
    {ARM7opCMNSimm     ,1},
    {ARM7opCMNSimm     ,1},

    /*0x38*/
    {ARM7opORRimm      ,1},
    {ARM7opORRimm      ,1},
    {ARM7opORRimm      ,1},
    {ARM7opORRimm      ,1},
    {ARM7opORRimm      ,1},
    {ARM7opORRimm      ,1},
    {ARM7opORRimm      ,1},
    {ARM7opORRimm      ,1},
    {ARM7opORRimm      ,1},
    {ARM7opORRimm      ,1},
    {ARM7opORRimm      ,1},
    {ARM7opORRimm      ,1},
    {ARM7opORRimm      ,1},
    {ARM7opORRimm      ,1},
    {ARM7opORRimm      ,1},
    {ARM7opORRimm      ,1},

    /*0x39*/
    {ARM7opORRSimm     ,1},
    {ARM7opORRSimm     ,1},
    {ARM7opORRSimm     ,1},
    {ARM7opORRSimm     ,1},
    {ARM7opORRSimm     ,1},
    {ARM7opORRSimm     ,1},
    {ARM7opORRSimm     ,1},
    {ARM7opORRSimm     ,1},
    {ARM7opORRSimm     ,1},
    {ARM7opORRSimm     ,1},
    {ARM7opORRSimm     ,1},
    {ARM7opORRSimm     ,1},
    {ARM7opORRSimm     ,1},
    {ARM7opORRSimm     ,1},
    {ARM7opORRSimm     ,1},
    {ARM7opORRSimm     ,1},

    /*0x3A*/
    {ARM7opMOVimm      ,1},
    {ARM7opMOVimm      ,1},
    {ARM7opMOVimm      ,1},
    {ARM7opMOVimm      ,1},
    {ARM7opMOVimm      ,1},
    {ARM7opMOVimm      ,1},
    {ARM7opMOVimm      ,1},
    {ARM7opMOVimm      ,1},
    {ARM7opMOVimm      ,1},
    {ARM7opMOVimm      ,1},
    {ARM7opMOVimm      ,1},
    {ARM7opMOVimm      ,1},
    {ARM7opMOVimm      ,1},
    {ARM7opMOVimm      ,1},
    {ARM7opMOVimm      ,1},
    {ARM7opMOVimm      ,1},

    /*0x3B*/
    {ARM7opMOVSimm     ,1},
    {ARM7opMOVSimm     ,1},
    {ARM7opMOVSimm     ,1},
    {ARM7opMOVSimm     ,1},
    {ARM7opMOVSimm     ,1},
    {ARM7opMOVSimm     ,1},
    {ARM7opMOVSimm     ,1},
    {ARM7opMOVSimm     ,1},
    {ARM7opMOVSimm     ,1},
    {ARM7opMOVSimm     ,1},
    {ARM7opMOVSimm     ,1},
    {ARM7opMOVSimm     ,1},
    {ARM7opMOVSimm     ,1},
    {ARM7opMOVSimm     ,1},
    {ARM7opMOVSimm     ,1},
    {ARM7opMOVSimm     ,1},

    /*0x3C*/
    {ARM7opBICimm      ,1},
    {ARM7opBICimm      ,1},
    {ARM7opBICimm      ,1},
    {ARM7opBICimm      ,1},
    {ARM7opBICimm      ,1},
    {ARM7opBICimm      ,1},
    {ARM7opBICimm      ,1},
    {ARM7opBICimm      ,1},
    {ARM7opBICimm      ,1},
    {ARM7opBICimm      ,1},
    {ARM7opBICimm      ,1},
    {ARM7opBICimm      ,1},
    {ARM7opBICimm      ,1},
    {ARM7opBICimm      ,1},
    {ARM7opBICimm      ,1},
    {ARM7opBICimm      ,1},

    /*0x3D*/
    {ARM7opBICSimm     ,1},
    {ARM7opBICSimm     ,1},
    {ARM7opBICSimm     ,1},
    {ARM7opBICSimm     ,1},
    {ARM7opBICSimm     ,1},
    {ARM7opBICSimm     ,1},
    {ARM7opBICSimm     ,1},
    {ARM7opBICSimm     ,1},
    {ARM7opBICSimm     ,1},
    {ARM7opBICSimm     ,1},
    {ARM7opBICSimm     ,1},
    {ARM7opBICSimm     ,1},
    {ARM7opBICSimm     ,1},
    {ARM7opBICSimm     ,1},
    {ARM7opBICSimm     ,1},
    {ARM7opBICSimm     ,1},

    /*0x3E*/
    {ARM7opMVNimm      ,1},
    {ARM7opMVNimm      ,1},
    {ARM7opMVNimm      ,1},
    {ARM7opMVNimm      ,1},
    {ARM7opMVNimm      ,1},
    {ARM7opMVNimm      ,1},
    {ARM7opMVNimm      ,1},
    {ARM7opMVNimm      ,1},
    {ARM7opMVNimm      ,1},
    {ARM7opMVNimm      ,1},
    {ARM7opMVNimm      ,1},
    {ARM7opMVNimm      ,1},
    {ARM7opMVNimm      ,1},
    {ARM7opMVNimm      ,1},
    {ARM7opMVNimm      ,1},
    {ARM7opMVNimm      ,1},

    /*0x3F*/
    {ARM7opMVNSimm     ,1},
    {ARM7opMVNSimm     ,1},
    {ARM7opMVNSimm     ,1},
    {ARM7opMVNSimm     ,1},
    {ARM7opMVNSimm     ,1},
    {ARM7opMVNSimm     ,1},
    {ARM7opMVNSimm     ,1},
    {ARM7opMVNSimm     ,1},
    {ARM7opMVNSimm     ,1},
    {ARM7opMVNSimm     ,1},
    {ARM7opMVNSimm     ,1},
    {ARM7opMVNSimm     ,1},
    {ARM7opMVNSimm     ,1},
    {ARM7opMVNSimm     ,1},
    {ARM7opMVNSimm     ,1},
    {ARM7opMVNSimm     ,1},

    /*0x40*/
    {ARM7opSTRptim     ,2},
    {ARM7opSTRptim     ,2},
    {ARM7opSTRptim     ,2},
    {ARM7opSTRptim     ,2},
    {ARM7opSTRptim     ,2},
    {ARM7opSTRptim     ,2},
    {ARM7opSTRptim     ,2},
    {ARM7opSTRptim     ,2},
    {ARM7opSTRptim     ,2},
    {ARM7opSTRptim     ,2},
    {ARM7opSTRptim     ,2},
    {ARM7opSTRptim     ,2},
    {ARM7opSTRptim     ,2},
    {ARM7opSTRptim     ,2},
    {ARM7opSTRptim     ,2},
    {ARM7opSTRptim     ,2},

    /*0x41*/
    {ARM7opLDRptim     ,3},
    {ARM7opLDRptim     ,3},
    {ARM7opLDRptim     ,3},
    {ARM7opLDRptim     ,3},
    {ARM7opLDRptim     ,3},
    {ARM7opLDRptim     ,3},
    {ARM7opLDRptim     ,3},
    {ARM7opLDRptim     ,3},
    {ARM7opLDRptim     ,3},
    {ARM7opLDRptim     ,3},
    {ARM7opLDRptim     ,3},
    {ARM7opLDRptim     ,3},
    {ARM7opLDRptim     ,3},
    {ARM7opLDRptim     ,3},
    {ARM7opLDRptim     ,3},
    {ARM7opLDRptim     ,3},

    /*0x42*/
    {ARM7opSTRTptim    ,2},
    {ARM7opSTRTptim    ,2},
    {ARM7opSTRTptim    ,2},
    {ARM7opSTRTptim    ,2},
    {ARM7opSTRTptim    ,2},
    {ARM7opSTRTptim    ,2},
    {ARM7opSTRTptim    ,2},
    {ARM7opSTRTptim    ,2},
    {ARM7opSTRTptim    ,2},
    {ARM7opSTRTptim    ,2},
    {ARM7opSTRTptim    ,2},
    {ARM7opSTRTptim    ,2},
    {ARM7opSTRTptim    ,2},
    {ARM7opSTRTptim    ,2},
    {ARM7opSTRTptim    ,2},
    {ARM7opSTRTptim    ,2},

    /*0x43*/
    {ARM7opLDRTptim    ,3},
    {ARM7opLDRTptim    ,3},
    {ARM7opLDRTptim    ,3},
    {ARM7opLDRTptim    ,3},
    {ARM7opLDRTptim    ,3},
    {ARM7opLDRTptim    ,3},
    {ARM7opLDRTptim    ,3},
    {ARM7opLDRTptim    ,3},
    {ARM7opLDRTptim    ,3},
    {ARM7opLDRTptim    ,3},
    {ARM7opLDRTptim    ,3},
    {ARM7opLDRTptim    ,3},
    {ARM7opLDRTptim    ,3},
    {ARM7opLDRTptim    ,3},
    {ARM7opLDRTptim    ,3},
    {ARM7opLDRTptim    ,3},

    /*0x44*/
    {ARM7opSTRBptim    ,2},
    {ARM7opSTRBptim    ,2},
    {ARM7opSTRBptim    ,2},
    {ARM7opSTRBptim    ,2},
    {ARM7opSTRBptim    ,2},
    {ARM7opSTRBptim    ,2},
    {ARM7opSTRBptim    ,2},
    {ARM7opSTRBptim    ,2},
    {ARM7opSTRBptim    ,2},
    {ARM7opSTRBptim    ,2},
    {ARM7opSTRBptim    ,2},
    {ARM7opSTRBptim    ,2},
    {ARM7opSTRBptim    ,2},
    {ARM7opSTRBptim    ,2},
    {ARM7opSTRBptim    ,2},
    {ARM7opSTRBptim    ,2},

    /*0x45*/
    {ARM7opLDRBptim    ,3},
    {ARM7opLDRBptim    ,3},
    {ARM7opLDRBptim    ,3},
    {ARM7opLDRBptim    ,3},
    {ARM7opLDRBptim    ,3},
    {ARM7opLDRBptim    ,3},
    {ARM7opLDRBptim    ,3},
    {ARM7opLDRBptim    ,3},
    {ARM7opLDRBptim    ,3},
    {ARM7opLDRBptim    ,3},
    {ARM7opLDRBptim    ,3},
    {ARM7opLDRBptim    ,3},
    {ARM7opLDRBptim    ,3},
    {ARM7opLDRBptim    ,3},
    {ARM7opLDRBptim    ,3},
    {ARM7opLDRBptim    ,3},

    /*0x46*/
    {ARM7opSTRBTptim   ,2},
    {ARM7opSTRBTptim   ,2},
    {ARM7opSTRBTptim   ,2},
    {ARM7opSTRBTptim   ,2},
    {ARM7opSTRBTptim   ,2},
    {ARM7opSTRBTptim   ,2},
    {ARM7opSTRBTptim   ,2},
    {ARM7opSTRBTptim   ,2},
    {ARM7opSTRBTptim   ,2},
    {ARM7opSTRBTptim   ,2},
    {ARM7opSTRBTptim   ,2},
    {ARM7opSTRBTptim   ,2},
    {ARM7opSTRBTptim   ,2},
    {ARM7opSTRBTptim   ,2},
    {ARM7opSTRBTptim   ,2},
    {ARM7opSTRBTptim   ,2},

    /*0x47*/
    {ARM7opLDRBTptim   ,3},
    {ARM7opLDRBTptim   ,3},
    {ARM7opLDRBTptim   ,3},
    {ARM7opLDRBTptim   ,3},
    {ARM7opLDRBTptim   ,3},
    {ARM7opLDRBTptim   ,3},
    {ARM7opLDRBTptim   ,3},
    {ARM7opLDRBTptim   ,3},
    {ARM7opLDRBTptim   ,3},
    {ARM7opLDRBTptim   ,3},
    {ARM7opLDRBTptim   ,3},
    {ARM7opLDRBTptim   ,3},
    {ARM7opLDRBTptim   ,3},
    {ARM7opLDRBTptim   ,3},
    {ARM7opLDRBTptim   ,3},
    {ARM7opLDRBTptim   ,3},

    /*0x48*/
    {ARM7opSTRptip     ,2},
    {ARM7opSTRptip     ,2},
    {ARM7opSTRptip     ,2},
    {ARM7opSTRptip     ,2},
    {ARM7opSTRptip     ,2},
    {ARM7opSTRptip     ,2},
    {ARM7opSTRptip     ,2},
    {ARM7opSTRptip     ,2},
    {ARM7opSTRptip     ,2},
    {ARM7opSTRptip     ,2},
    {ARM7opSTRptip     ,2},
    {ARM7opSTRptip     ,2},
    {ARM7opSTRptip     ,2},
    {ARM7opSTRptip     ,2},
    {ARM7opSTRptip     ,2},
    {ARM7opSTRptip     ,2},

    /*0x49*/
    {ARM7opLDRptip     ,3},
    {ARM7opLDRptip     ,3},
    {ARM7opLDRptip     ,3},
    {ARM7opLDRptip     ,3},
    {ARM7opLDRptip     ,3},
    {ARM7opLDRptip     ,3},
    {ARM7opLDRptip     ,3},
    {ARM7opLDRptip     ,3},
    {ARM7opLDRptip     ,3},
    {ARM7opLDRptip     ,3},
    {ARM7opLDRptip     ,3},
    {ARM7opLDRptip     ,3},
    {ARM7opLDRptip     ,3},
    {ARM7opLDRptip     ,3},
    {ARM7opLDRptip     ,3},
    {ARM7opLDRptip     ,3},

    /*0x4A*/
    {ARM7opSTRTptip    ,2},
    {ARM7opSTRTptip    ,2},
    {ARM7opSTRTptip    ,2},
    {ARM7opSTRTptip    ,2},
    {ARM7opSTRTptip    ,2},
    {ARM7opSTRTptip    ,2},
    {ARM7opSTRTptip    ,2},
    {ARM7opSTRTptip    ,2},
    {ARM7opSTRTptip    ,2},
    {ARM7opSTRTptip    ,2},
    {ARM7opSTRTptip    ,2},
    {ARM7opSTRTptip    ,2},
    {ARM7opSTRTptip    ,2},
    {ARM7opSTRTptip    ,2},
    {ARM7opSTRTptip    ,2},
    {ARM7opSTRTptip    ,2},

    /*0x4B*/
    {ARM7opLDRTptip    ,3},
    {ARM7opLDRTptip    ,3},
    {ARM7opLDRTptip    ,3},
    {ARM7opLDRTptip    ,3},
    {ARM7opLDRTptip    ,3},
    {ARM7opLDRTptip    ,3},
    {ARM7opLDRTptip    ,3},
    {ARM7opLDRTptip    ,3},
    {ARM7opLDRTptip    ,3},
    {ARM7opLDRTptip    ,3},
    {ARM7opLDRTptip    ,3},
    {ARM7opLDRTptip    ,3},
    {ARM7opLDRTptip    ,3},
    {ARM7opLDRTptip    ,3},
    {ARM7opLDRTptip    ,3},
    {ARM7opLDRTptip    ,3},

    /*0x4C*/
    {ARM7opSTRBptip    ,2},
    {ARM7opSTRBptip    ,2},
    {ARM7opSTRBptip    ,2},
    {ARM7opSTRBptip    ,2},
    {ARM7opSTRBptip    ,2},
    {ARM7opSTRBptip    ,2},
    {ARM7opSTRBptip    ,2},
    {ARM7opSTRBptip    ,2},
    {ARM7opSTRBptip    ,2},
    {ARM7opSTRBptip    ,2},
    {ARM7opSTRBptip    ,2},
    {ARM7opSTRBptip    ,2},
    {ARM7opSTRBptip    ,2},
    {ARM7opSTRBptip    ,2},
    {ARM7opSTRBptip    ,2},
    {ARM7opSTRBptip    ,2},

    /*0x4D*/
    {ARM7opLDRBptip    ,3},
    {ARM7opLDRBptip    ,3},
    {ARM7opLDRBptip    ,3},
    {ARM7opLDRBptip    ,3},
    {ARM7opLDRBptip    ,3},
    {ARM7opLDRBptip    ,3},
    {ARM7opLDRBptip    ,3},
    {ARM7opLDRBptip    ,3},
    {ARM7opLDRBptip    ,3},
    {ARM7opLDRBptip    ,3},
    {ARM7opLDRBptip    ,3},
    {ARM7opLDRBptip    ,3},
    {ARM7opLDRBptip    ,3},
    {ARM7opLDRBptip    ,3},
    {ARM7opLDRBptip    ,3},
    {ARM7opLDRBptip    ,3},

    /*0x4E*/
    {ARM7opSTRBTptip   ,2},
    {ARM7opSTRBTptip   ,2},
    {ARM7opSTRBTptip   ,2},
    {ARM7opSTRBTptip   ,2},
    {ARM7opSTRBTptip   ,2},
    {ARM7opSTRBTptip   ,2},
    {ARM7opSTRBTptip   ,2},
    {ARM7opSTRBTptip   ,2},
    {ARM7opSTRBTptip   ,2},
    {ARM7opSTRBTptip   ,2},
    {ARM7opSTRBTptip   ,2},
    {ARM7opSTRBTptip   ,2},
    {ARM7opSTRBTptip   ,2},
    {ARM7opSTRBTptip   ,2},
    {ARM7opSTRBTptip   ,2},
    {ARM7opSTRBTptip   ,2},

    /*0x4F*/
    {ARM7opLDRBTptip   ,3},
    {ARM7opLDRBTptip   ,3},
    {ARM7opLDRBTptip   ,3},
    {ARM7opLDRBTptip   ,3},
    {ARM7opLDRBTptip   ,3},
    {ARM7opLDRBTptip   ,3},
    {ARM7opLDRBTptip   ,3},
    {ARM7opLDRBTptip   ,3},
    {ARM7opLDRBTptip   ,3},
    {ARM7opLDRBTptip   ,3},
    {ARM7opLDRBTptip   ,3},
    {ARM7opLDRBTptip   ,3},
    {ARM7opLDRBTptip   ,3},
    {ARM7opLDRBTptip   ,3},
    {ARM7opLDRBTptip   ,3},
    {ARM7opLDRBTptip   ,3},

    /*0x50*/
    {ARM7opSTRofim     ,2},
    {ARM7opSTRofim     ,2},
    {ARM7opSTRofim     ,2},
    {ARM7opSTRofim     ,2},
    {ARM7opSTRofim     ,2},
    {ARM7opSTRofim     ,2},
    {ARM7opSTRofim     ,2},
    {ARM7opSTRofim     ,2},
    {ARM7opSTRofim     ,2},
    {ARM7opSTRofim     ,2},
    {ARM7opSTRofim     ,2},
    {ARM7opSTRofim     ,2},
    {ARM7opSTRofim     ,2},
    {ARM7opSTRofim     ,2},
    {ARM7opSTRofim     ,2},
    {ARM7opSTRofim     ,2},

    /*0x51*/
    {ARM7opLDRofim     ,3},
    {ARM7opLDRofim     ,3},
    {ARM7opLDRofim     ,3},
    {ARM7opLDRofim     ,3},
    {ARM7opLDRofim     ,3},
    {ARM7opLDRofim     ,3},
    {ARM7opLDRofim     ,3},
    {ARM7opLDRofim     ,3},
    {ARM7opLDRofim     ,3},
    {ARM7opLDRofim     ,3},
    {ARM7opLDRofim     ,3},
    {ARM7opLDRofim     ,3},
    {ARM7opLDRofim     ,3},
    {ARM7opLDRofim     ,3},
    {ARM7opLDRofim     ,3},
    {ARM7opLDRofim     ,3},

    /*0x52*/
    {ARM7opSTRprim     ,2},
    {ARM7opSTRprim     ,2},
    {ARM7opSTRprim     ,2},
    {ARM7opSTRprim     ,2},
    {ARM7opSTRprim     ,2},
    {ARM7opSTRprim     ,2},
    {ARM7opSTRprim     ,2},
    {ARM7opSTRprim     ,2},
    {ARM7opSTRprim     ,2},
    {ARM7opSTRprim     ,2},
    {ARM7opSTRprim     ,2},
    {ARM7opSTRprim     ,2},
    {ARM7opSTRprim     ,2},
    {ARM7opSTRprim     ,2},
    {ARM7opSTRprim     ,2},
    {ARM7opSTRprim     ,2},

    /*0x53*/
    {ARM7opLDRprim     ,3},
    {ARM7opLDRprim     ,3},
    {ARM7opLDRprim     ,3},
    {ARM7opLDRprim     ,3},
    {ARM7opLDRprim     ,3},
    {ARM7opLDRprim     ,3},
    {ARM7opLDRprim     ,3},
    {ARM7opLDRprim     ,3},
    {ARM7opLDRprim     ,3},
    {ARM7opLDRprim     ,3},
    {ARM7opLDRprim     ,3},
    {ARM7opLDRprim     ,3},
    {ARM7opLDRprim     ,3},
    {ARM7opLDRprim     ,3},
    {ARM7opLDRprim     ,3},
    {ARM7opLDRprim     ,3},

    /*0x54*/
    {ARM7opSTRBofim    ,2},
    {ARM7opSTRBofim    ,2},
    {ARM7opSTRBofim    ,2},
    {ARM7opSTRBofim    ,2},
    {ARM7opSTRBofim    ,2},
    {ARM7opSTRBofim    ,2},
    {ARM7opSTRBofim    ,2},
    {ARM7opSTRBofim    ,2},
    {ARM7opSTRBofim    ,2},
    {ARM7opSTRBofim    ,2},
    {ARM7opSTRBofim    ,2},
    {ARM7opSTRBofim    ,2},
    {ARM7opSTRBofim    ,2},
    {ARM7opSTRBofim    ,2},
    {ARM7opSTRBofim    ,2},
    {ARM7opSTRBofim    ,2},

    /*0x55*/
    {ARM7opLDRBofim    ,3},
    {ARM7opLDRBofim    ,3},
    {ARM7opLDRBofim    ,3},
    {ARM7opLDRBofim    ,3},
    {ARM7opLDRBofim    ,3},
    {ARM7opLDRBofim    ,3},
    {ARM7opLDRBofim    ,3},
    {ARM7opLDRBofim    ,3},
    {ARM7opLDRBofim    ,3},
    {ARM7opLDRBofim    ,3},
    {ARM7opLDRBofim    ,3},
    {ARM7opLDRBofim    ,3},
    {ARM7opLDRBofim    ,3},
    {ARM7opLDRBofim    ,3},
    {ARM7opLDRBofim    ,3},
    {ARM7opLDRBofim    ,3},

    /*0x56*/
    {ARM7opSTRBprim    ,2},
    {ARM7opSTRBprim    ,2},
    {ARM7opSTRBprim    ,2},
    {ARM7opSTRBprim    ,2},
    {ARM7opSTRBprim    ,2},
    {ARM7opSTRBprim    ,2},
    {ARM7opSTRBprim    ,2},
    {ARM7opSTRBprim    ,2},
    {ARM7opSTRBprim    ,2},
    {ARM7opSTRBprim    ,2},
    {ARM7opSTRBprim    ,2},
    {ARM7opSTRBprim    ,2},
    {ARM7opSTRBprim    ,2},
    {ARM7opSTRBprim    ,2},
    {ARM7opSTRBprim    ,2},
    {ARM7opSTRBprim    ,2},

    /*0x57*/
    {ARM7opLDRBprim    ,3},
    {ARM7opLDRBprim    ,3},
    {ARM7opLDRBprim    ,3},
    {ARM7opLDRBprim    ,3},
    {ARM7opLDRBprim    ,3},
    {ARM7opLDRBprim    ,3},
    {ARM7opLDRBprim    ,3},
    {ARM7opLDRBprim    ,3},
    {ARM7opLDRBprim    ,3},
    {ARM7opLDRBprim    ,3},
    {ARM7opLDRBprim    ,3},
    {ARM7opLDRBprim    ,3},
    {ARM7opLDRBprim    ,3},
    {ARM7opLDRBprim    ,3},
    {ARM7opLDRBprim    ,3},
    {ARM7opLDRBprim    ,3},

    /*0x58*/
    {ARM7opSTRofip     ,2},
    {ARM7opSTRofip     ,2},
    {ARM7opSTRofip     ,2},
    {ARM7opSTRofip     ,2},
    {ARM7opSTRofip     ,2},
    {ARM7opSTRofip     ,2},
    {ARM7opSTRofip     ,2},
    {ARM7opSTRofip     ,2},
    {ARM7opSTRofip     ,2},
    {ARM7opSTRofip     ,2},
    {ARM7opSTRofip     ,2},
    {ARM7opSTRofip     ,2},
    {ARM7opSTRofip     ,2},
    {ARM7opSTRofip     ,2},
    {ARM7opSTRofip     ,2},
    {ARM7opSTRofip     ,2},

    /*0x59*/
    {ARM7opLDRofip     ,3},
    {ARM7opLDRofip     ,3},
    {ARM7opLDRofip     ,3},
    {ARM7opLDRofip     ,3},
    {ARM7opLDRofip     ,3},
    {ARM7opLDRofip     ,3},
    {ARM7opLDRofip     ,3},
    {ARM7opLDRofip     ,3},
    {ARM7opLDRofip     ,3},
    {ARM7opLDRofip     ,3},
    {ARM7opLDRofip     ,3},
    {ARM7opLDRofip     ,3},
    {ARM7opLDRofip     ,3},
    {ARM7opLDRofip     ,3},
    {ARM7opLDRofip     ,3},
    {ARM7opLDRofip     ,3},

    /*0x5A*/
    {ARM7opSTRprip     ,2},
    {ARM7opSTRprip     ,2},
    {ARM7opSTRprip     ,2},
    {ARM7opSTRprip     ,2},
    {ARM7opSTRprip     ,2},
    {ARM7opSTRprip     ,2},
    {ARM7opSTRprip     ,2},
    {ARM7opSTRprip     ,2},
    {ARM7opSTRprip     ,2},
    {ARM7opSTRprip     ,2},
    {ARM7opSTRprip     ,2},
    {ARM7opSTRprip     ,2},
    {ARM7opSTRprip     ,2},
    {ARM7opSTRprip     ,2},
    {ARM7opSTRprip     ,2},
    {ARM7opSTRprip     ,2},

    /*0x5B*/
    {ARM7opLDRprip     ,3},
    {ARM7opLDRprip     ,3},
    {ARM7opLDRprip     ,3},
    {ARM7opLDRprip     ,3},
    {ARM7opLDRprip     ,3},
    {ARM7opLDRprip     ,3},
    {ARM7opLDRprip     ,3},
    {ARM7opLDRprip     ,3},
    {ARM7opLDRprip     ,3},
    {ARM7opLDRprip     ,3},
    {ARM7opLDRprip     ,3},
    {ARM7opLDRprip     ,3},
    {ARM7opLDRprip     ,3},
    {ARM7opLDRprip     ,3},
    {ARM7opLDRprip     ,3},
    {ARM7opLDRprip     ,3},

    /*0x5C*/
    {ARM7opSTRBofip    ,2},
    {ARM7opSTRBofip    ,2},
    {ARM7opSTRBofip    ,2},
    {ARM7opSTRBofip    ,2},
    {ARM7opSTRBofip    ,2},
    {ARM7opSTRBofip    ,2},
    {ARM7opSTRBofip    ,2},
    {ARM7opSTRBofip    ,2},
    {ARM7opSTRBofip    ,2},
    {ARM7opSTRBofip    ,2},
    {ARM7opSTRBofip    ,2},
    {ARM7opSTRBofip    ,2},
    {ARM7opSTRBofip    ,2},
    {ARM7opSTRBofip    ,2},
    {ARM7opSTRBofip    ,2},
    {ARM7opSTRBofip    ,2},

    /*0x5D*/
    {ARM7opLDRBofip    ,3},
    {ARM7opLDRBofip    ,3},
    {ARM7opLDRBofip    ,3},
    {ARM7opLDRBofip    ,3},
    {ARM7opLDRBofip    ,3},
    {ARM7opLDRBofip    ,3},
    {ARM7opLDRBofip    ,3},
    {ARM7opLDRBofip    ,3},
    {ARM7opLDRBofip    ,3},
    {ARM7opLDRBofip    ,3},
    {ARM7opLDRBofip    ,3},
    {ARM7opLDRBofip    ,3},
    {ARM7opLDRBofip    ,3},
    {ARM7opLDRBofip    ,3},
    {ARM7opLDRBofip    ,3},
    {ARM7opLDRBofip    ,3},

    /*0x5E*/
    {ARM7opSTRBprip    ,2},
    {ARM7opSTRBprip    ,2},
    {ARM7opSTRBprip    ,2},
    {ARM7opSTRBprip    ,2},
    {ARM7opSTRBprip    ,2},
    {ARM7opSTRBprip    ,2},
    {ARM7opSTRBprip    ,2},
    {ARM7opSTRBprip    ,2},
    {ARM7opSTRBprip    ,2},
    {ARM7opSTRBprip    ,2},
    {ARM7opSTRBprip    ,2},
    {ARM7opSTRBprip    ,2},
    {ARM7opSTRBprip    ,2},
    {ARM7opSTRBprip    ,2},
    {ARM7opSTRBprip    ,2},
    {ARM7opSTRBprip    ,2},

    /*0x5F*/
    {ARM7opLDRBprip    ,3},
    {ARM7opLDRBprip    ,3},
    {ARM7opLDRBprip    ,3},
    {ARM7opLDRBprip    ,3},
    {ARM7opLDRBprip    ,3},
    {ARM7opLDRBprip    ,3},
    {ARM7opLDRBprip    ,3},
    {ARM7opLDRBprip    ,3},
    {ARM7opLDRBprip    ,3},
    {ARM7opLDRBprip    ,3},
    {ARM7opLDRBprip    ,3},
    {ARM7opLDRBprip    ,3},
    {ARM7opLDRBprip    ,3},
    {ARM7opLDRBprip    ,3},
    {ARM7opLDRBprip    ,3},
    {ARM7opLDRBprip    ,3},

    /*0x60*/
    {ARM7opSTRptrmll   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRptrmlr   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRptrmar   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRptrmrr   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRptrmll   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRptrmlr   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRptrmar   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRptrmrr   ,2},
    {ARM7opUND         ,4},

    /*0x61*/
    {ARM7opLDRptrmll   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRptrmlr   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRptrmar   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRptrmrr   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRptrmll   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRptrmlr   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRptrmar   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRptrmrr   ,3},
    {ARM7opUND         ,4},

    /*0x62*/
    {ARM7opSTRTptrmll  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRTptrmlr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRTptrmar  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRTptrmrr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRTptrmll  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRTptrmlr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRTptrmar  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRTptrmrr  ,2},
    {ARM7opUND         ,4},

    /*0x63*/
    {ARM7opLDRTptrmll  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRTptrmlr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRTptrmar  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRTptrmrr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRTptrmll  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRTptrmlr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRTptrmar  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRTptrmrr  ,3},
    {ARM7opUND         ,4},

    /*0x64*/
    {ARM7opSTRBptrmll  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBptrmlr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBptrmar  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBptrmrr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBptrmll  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBptrmlr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBptrmar  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBptrmrr  ,2},
    {ARM7opUND         ,4},

    /*0x65*/
    {ARM7opLDRBptrmll  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBptrmlr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBptrmar  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBptrmrr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBptrmll  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBptrmlr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBptrmar  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBptrmrr  ,3},
    {ARM7opUND         ,4},

    /*0x66*/
    {ARM7opSTRBTptrmll ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBTptrmlr ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBTptrmar ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBTptrmrr ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBTptrmll ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBTptrmlr ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBTptrmar ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBTptrmrr ,2},
    {ARM7opUND         ,4},

    /*0x67*/
    {ARM7opLDRBTptrmll ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBTptrmlr ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBTptrmar ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBTptrmrr ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBTptrmll ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBTptrmlr ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBTptrmar ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBTptrmrr ,3},
    {ARM7opUND         ,4},

    /*0x68*/
    {ARM7opSTRptrpll   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRptrplr   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRptrpar   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRptrprr   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRptrpll   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRptrplr   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRptrpar   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRptrprr   ,2},
    {ARM7opUND         ,4},

    /*0x69*/
    {ARM7opLDRptrpll   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRptrplr   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRptrpar   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRptrprr   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRptrpll   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRptrplr   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRptrpar   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRptrprr   ,3},
    {ARM7opUND         ,4},

    /*0x6A*/
    {ARM7opSTRTptrpll  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRTptrplr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRTptrpar  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRTptrprr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRTptrpll  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRTptrplr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRTptrpar  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRTptrprr  ,2},
    {ARM7opUND         ,4},

    /*0x6B*/
    {ARM7opLDRTptrpll  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRTptrplr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRTptrpar  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRTptrprr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRTptrpll  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRTptrplr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRTptrpar  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRTptrprr  ,3},
    {ARM7opUND         ,4},

    /*0x6C*/
    {ARM7opSTRBptrpll  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBptrplr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBptrpar  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBptrprr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBptrpll  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBptrplr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBptrpar  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBptrprr  ,2},
    {ARM7opUND         ,4},

    /*0x6D*/
    {ARM7opLDRBptrpll  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBptrplr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBptrpar  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBptrprr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBptrpll  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBptrplr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBptrpar  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBptrprr  ,3},
    {ARM7opUND         ,4},

    /*0x6E*/
    {ARM7opSTRBTptrpll ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBTptrplr ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBTptrpar ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBTptrprr ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBTptrpll ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBTptrplr ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBTptrpar ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBTptrprr ,2},
    {ARM7opUND         ,4},

    /*0x6F*/
    {ARM7opLDRBTptrpll ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBTptrplr ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBTptrpar ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBTptrprr ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBTptrpll ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBTptrplr ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBTptrpar ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBTptrprr ,3},
    {ARM7opUND         ,4},

    /*0x70*/
    {ARM7opSTRofrmll   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRofrmlr   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRofrmar   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRofrmrr   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRofrmll   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRofrmlr   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRofrmar   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRofrmrr   ,2},
    {ARM7opUND         ,4},

    /*0x71*/
    {ARM7opLDRofrmll   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRofrmlr   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRofrmar   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRofrmrr   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRofrmll   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRofrmlr   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRofrmar   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRofrmrr   ,3},
    {ARM7opUND         ,4},

    /*0x72*/
    {ARM7opSTRprrmll   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRprrmlr   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRprrmar   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRprrmrr   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRprrmll   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRprrmlr   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRprrmar   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRprrmrr   ,2},
    {ARM7opUND         ,4},

    /*0x73*/
    {ARM7opLDRprrmll   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRprrmlr   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRprrmar   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRprrmrr   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRprrmll   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRprrmlr   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRprrmar   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRprrmrr   ,3},
    {ARM7opUND         ,4},

    /*0x74*/
    {ARM7opSTRBofrmll  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBofrmlr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBofrmar  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBofrmrr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBofrmll  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBofrmlr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBofrmar  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBofrmrr  ,2},
    {ARM7opUND         ,4},

    /*0x75*/
    {ARM7opLDRBofrmll  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBofrmlr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBofrmar  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBofrmrr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBofrmll  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBofrmlr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBofrmar  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBofrmrr  ,3},
    {ARM7opUND         ,4},

    /*0x76*/
    {ARM7opSTRBprrmll  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBprrmlr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBprrmar  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBprrmrr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBprrmll  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBprrmlr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBprrmar  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBprrmrr  ,2},
    {ARM7opUND         ,4},

    /*0x77*/
    {ARM7opLDRBprrmll  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBprrmlr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBprrmar  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBprrmrr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBprrmll  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBprrmlr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBprrmar  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBprrmrr  ,3},
    {ARM7opUND         ,4},

    /*0x78*/
    {ARM7opSTRofrpll   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRofrplr   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRofrpar   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRofrprr   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRofrpll   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRofrplr   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRofrpar   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRofrprr   ,2},
    {ARM7opUND         ,4},

    /*0x79*/
    {ARM7opLDRofrpll   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRofrplr   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRofrpar   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRofrprr   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRofrpll   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRofrplr   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRofrpar   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRofrprr   ,3},
    {ARM7opUND         ,4},

    /*0x7A*/
    {ARM7opSTRprrpll   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRprrplr   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRprrpar   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRprrprr   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRprrpll   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRprrplr   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRprrpar   ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRprrprr   ,2},
    {ARM7opUND         ,4},

    /*0x7B*/
    {ARM7opLDRprrpll   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRprrplr   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRprrpar   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRprrprr   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRprrpll   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRprrplr   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRprrpar   ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRprrprr   ,3},
    {ARM7opUND         ,4},

    /*0x7C*/
    {ARM7opSTRBofrpll  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBofrplr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBofrpar  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBofrprr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBofrpll  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBofrplr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBofrpar  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBofrprr  ,2},
    {ARM7opUND         ,4},

    /*0x7D*/
    {ARM7opLDRBofrpll  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBofrplr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBofrpar  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBofrprr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBofrpll  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBofrplr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBofrpar  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBofrprr  ,3},
    {ARM7opUND         ,4},

    /*0x7E*/
    {ARM7opSTRBprrpll  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBprrplr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBprrpar  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBprrprr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBprrpll  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBprrplr  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBprrpar  ,2},
    {ARM7opUND         ,4},
    {ARM7opSTRBprrprr  ,2},
    {ARM7opUND         ,4},

    /*0x7F*/
    {ARM7opLDRBprrpll  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBprrplr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBprrpar  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBprrprr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBprrpll  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBprrplr  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBprrpar  ,3},
    {ARM7opUND         ,4},
    {ARM7opLDRBprrprr  ,3},
    {ARM7opUND         ,4},

    /*0x80*/
    {ARM7opSTMDA       ,2},
    {ARM7opSTMDA       ,2},
    {ARM7opSTMDA       ,2},
    {ARM7opSTMDA       ,2},
    {ARM7opSTMDA       ,2},
    {ARM7opSTMDA       ,2},
    {ARM7opSTMDA       ,2},
    {ARM7opSTMDA       ,2},
    {ARM7opSTMDA       ,2},
    {ARM7opSTMDA       ,2},
    {ARM7opSTMDA       ,2},
    {ARM7opSTMDA       ,2},
    {ARM7opSTMDA       ,2},
    {ARM7opSTMDA       ,2},
    {ARM7opSTMDA       ,2},
    {ARM7opSTMDA       ,2},

    /*0x81*/
    {ARM7opLDMDA       ,3},
    {ARM7opLDMDA       ,3},
    {ARM7opLDMDA       ,3},
    {ARM7opLDMDA       ,3},
    {ARM7opLDMDA       ,3},
    {ARM7opLDMDA       ,3},
    {ARM7opLDMDA       ,3},
    {ARM7opLDMDA       ,3},
    {ARM7opLDMDA       ,3},
    {ARM7opLDMDA       ,3},
    {ARM7opLDMDA       ,3},
    {ARM7opLDMDA       ,3},
    {ARM7opLDMDA       ,3},
    {ARM7opLDMDA       ,3},
    {ARM7opLDMDA       ,3},
    {ARM7opLDMDA       ,3},

    /*0x82*/
    {ARM7opSTMDAw      ,2},
    {ARM7opSTMDAw      ,2},
    {ARM7opSTMDAw      ,2},
    {ARM7opSTMDAw      ,2},
    {ARM7opSTMDAw      ,2},
    {ARM7opSTMDAw      ,2},
    {ARM7opSTMDAw      ,2},
    {ARM7opSTMDAw      ,2},
    {ARM7opSTMDAw      ,2},
    {ARM7opSTMDAw      ,2},
    {ARM7opSTMDAw      ,2},
    {ARM7opSTMDAw      ,2},
    {ARM7opSTMDAw      ,2},
    {ARM7opSTMDAw      ,2},
    {ARM7opSTMDAw      ,2},
    {ARM7opSTMDAw      ,2},

    /*0x83*/
    {ARM7opLDMDAw      ,3},
    {ARM7opLDMDAw      ,3},
    {ARM7opLDMDAw      ,3},
    {ARM7opLDMDAw      ,3},
    {ARM7opLDMDAw      ,3},
    {ARM7opLDMDAw      ,3},
    {ARM7opLDMDAw      ,3},
    {ARM7opLDMDAw      ,3},
    {ARM7opLDMDAw      ,3},
    {ARM7opLDMDAw      ,3},
    {ARM7opLDMDAw      ,3},
    {ARM7opLDMDAw      ,3},
    {ARM7opLDMDAw      ,3},
    {ARM7opLDMDAw      ,3},
    {ARM7opLDMDAw      ,3},
    {ARM7opLDMDAw      ,3},

    /*0x84*/
    {ARM7opSTMDAu      ,2},
    {ARM7opSTMDAu      ,2},
    {ARM7opSTMDAu      ,2},
    {ARM7opSTMDAu      ,2},
    {ARM7opSTMDAu      ,2},
    {ARM7opSTMDAu      ,2},
    {ARM7opSTMDAu      ,2},
    {ARM7opSTMDAu      ,2},
    {ARM7opSTMDAu      ,2},
    {ARM7opSTMDAu      ,2},
    {ARM7opSTMDAu      ,2},
    {ARM7opSTMDAu      ,2},
    {ARM7opSTMDAu      ,2},
    {ARM7opSTMDAu      ,2},
    {ARM7opSTMDAu      ,2},
    {ARM7opSTMDAu      ,2},

    /*0x85*/
    {ARM7opLDMDAu      ,3},
    {ARM7opLDMDAu      ,3},
    {ARM7opLDMDAu      ,3},
    {ARM7opLDMDAu      ,3},
    {ARM7opLDMDAu      ,3},
    {ARM7opLDMDAu      ,3},
    {ARM7opLDMDAu      ,3},
    {ARM7opLDMDAu      ,3},
    {ARM7opLDMDAu      ,3},
    {ARM7opLDMDAu      ,3},
    {ARM7opLDMDAu      ,3},
    {ARM7opLDMDAu      ,3},
    {ARM7opLDMDAu      ,3},
    {ARM7opLDMDAu      ,3},
    {ARM7opLDMDAu      ,3},
    {ARM7opLDMDAu      ,3},

    /*0x86*/
    {ARM7opSTMDAuw     ,2},
    {ARM7opSTMDAuw     ,2},
    {ARM7opSTMDAuw     ,2},
    {ARM7opSTMDAuw     ,2},
    {ARM7opSTMDAuw     ,2},
    {ARM7opSTMDAuw     ,2},
    {ARM7opSTMDAuw     ,2},
    {ARM7opSTMDAuw     ,2},
    {ARM7opSTMDAuw     ,2},
    {ARM7opSTMDAuw     ,2},
    {ARM7opSTMDAuw     ,2},
    {ARM7opSTMDAuw     ,2},
    {ARM7opSTMDAuw     ,2},
    {ARM7opSTMDAuw     ,2},
    {ARM7opSTMDAuw     ,2},
    {ARM7opSTMDAuw     ,2},

    /*0x87*/
    {ARM7opLDMDAuw     ,3},
    {ARM7opLDMDAuw     ,3},
    {ARM7opLDMDAuw     ,3},
    {ARM7opLDMDAuw     ,3},
    {ARM7opLDMDAuw     ,3},
    {ARM7opLDMDAuw     ,3},
    {ARM7opLDMDAuw     ,3},
    {ARM7opLDMDAuw     ,3},
    {ARM7opLDMDAuw     ,3},
    {ARM7opLDMDAuw     ,3},
    {ARM7opLDMDAuw     ,3},
    {ARM7opLDMDAuw     ,3},
    {ARM7opLDMDAuw     ,3},
    {ARM7opLDMDAuw     ,3},
    {ARM7opLDMDAuw     ,3},
    {ARM7opLDMDAuw     ,3},

    /*0x88*/
    {ARM7opSTMIA       ,2},
    {ARM7opSTMIA       ,2},
    {ARM7opSTMIA       ,2},
    {ARM7opSTMIA       ,2},
    {ARM7opSTMIA       ,2},
    {ARM7opSTMIA       ,2},
    {ARM7opSTMIA       ,2},
    {ARM7opSTMIA       ,2},
    {ARM7opSTMIA       ,2},
    {ARM7opSTMIA       ,2},
    {ARM7opSTMIA       ,2},
    {ARM7opSTMIA       ,2},
    {ARM7opSTMIA       ,2},
    {ARM7opSTMIA       ,2},
    {ARM7opSTMIA       ,2},
    {ARM7opSTMIA       ,2},

    /*0x89*/
    {ARM7opLDMIA       ,3},
    {ARM7opLDMIA       ,3},
    {ARM7opLDMIA       ,3},
    {ARM7opLDMIA       ,3},
    {ARM7opLDMIA       ,3},
    {ARM7opLDMIA       ,3},
    {ARM7opLDMIA       ,3},
    {ARM7opLDMIA       ,3},
    {ARM7opLDMIA       ,3},
    {ARM7opLDMIA       ,3},
    {ARM7opLDMIA       ,3},
    {ARM7opLDMIA       ,3},
    {ARM7opLDMIA       ,3},
    {ARM7opLDMIA       ,3},
    {ARM7opLDMIA       ,3},
    {ARM7opLDMIA       ,3},

    /*0x8A*/
    {ARM7opSTMIAw      ,2},
    {ARM7opSTMIAw      ,2},
    {ARM7opSTMIAw      ,2},
    {ARM7opSTMIAw      ,2},
    {ARM7opSTMIAw      ,2},
    {ARM7opSTMIAw      ,2},
    {ARM7opSTMIAw      ,2},
    {ARM7opSTMIAw      ,2},
    {ARM7opSTMIAw      ,2},
    {ARM7opSTMIAw      ,2},
    {ARM7opSTMIAw      ,2},
    {ARM7opSTMIAw      ,2},
    {ARM7opSTMIAw      ,2},
    {ARM7opSTMIAw      ,2},
    {ARM7opSTMIAw      ,2},
    {ARM7opSTMIAw      ,2},

    /*0x8B*/
    {ARM7opLDMIAw      ,3},
    {ARM7opLDMIAw      ,3},
    {ARM7opLDMIAw      ,3},
    {ARM7opLDMIAw      ,3},
    {ARM7opLDMIAw      ,3},
    {ARM7opLDMIAw      ,3},
    {ARM7opLDMIAw      ,3},
    {ARM7opLDMIAw      ,3},
    {ARM7opLDMIAw      ,3},
    {ARM7opLDMIAw      ,3},
    {ARM7opLDMIAw      ,3},
    {ARM7opLDMIAw      ,3},
    {ARM7opLDMIAw      ,3},
    {ARM7opLDMIAw      ,3},
    {ARM7opLDMIAw      ,3},
    {ARM7opLDMIAw      ,3},

    /*0x8C*/
    {ARM7opSTMIAu      ,2},
    {ARM7opSTMIAu      ,2},
    {ARM7opSTMIAu      ,2},
    {ARM7opSTMIAu      ,2},
    {ARM7opSTMIAu      ,2},
    {ARM7opSTMIAu      ,2},
    {ARM7opSTMIAu      ,2},
    {ARM7opSTMIAu      ,2},
    {ARM7opSTMIAu      ,2},
    {ARM7opSTMIAu      ,2},
    {ARM7opSTMIAu      ,2},
    {ARM7opSTMIAu      ,2},
    {ARM7opSTMIAu      ,2},
    {ARM7opSTMIAu      ,2},
    {ARM7opSTMIAu      ,2},
    {ARM7opSTMIAu      ,2},

    /*0x8D*/
    {ARM7opLDMIAu      ,3},
    {ARM7opLDMIAu      ,3},
    {ARM7opLDMIAu      ,3},
    {ARM7opLDMIAu      ,3},
    {ARM7opLDMIAu      ,3},
    {ARM7opLDMIAu      ,3},
    {ARM7opLDMIAu      ,3},
    {ARM7opLDMIAu      ,3},
    {ARM7opLDMIAu      ,3},
    {ARM7opLDMIAu      ,3},
    {ARM7opLDMIAu      ,3},
    {ARM7opLDMIAu      ,3},
    {ARM7opLDMIAu      ,3},
    {ARM7opLDMIAu      ,3},
    {ARM7opLDMIAu      ,3},
    {ARM7opLDMIAu      ,3},

    /*0x8E*/
    {ARM7opSTMIAuw     ,2},
    {ARM7opSTMIAuw     ,2},
    {ARM7opSTMIAuw     ,2},
    {ARM7opSTMIAuw     ,2},
    {ARM7opSTMIAuw     ,2},
    {ARM7opSTMIAuw     ,2},
    {ARM7opSTMIAuw     ,2},
    {ARM7opSTMIAuw     ,2},
    {ARM7opSTMIAuw     ,2},
    {ARM7opSTMIAuw     ,2},
    {ARM7opSTMIAuw     ,2},
    {ARM7opSTMIAuw     ,2},
    {ARM7opSTMIAuw     ,2},
    {ARM7opSTMIAuw     ,2},
    {ARM7opSTMIAuw     ,2},
    {ARM7opSTMIAuw     ,2},

    /*0x8F*/
    {ARM7opLDMIAuw     ,3},
    {ARM7opLDMIAuw     ,3},
    {ARM7opLDMIAuw     ,3},
    {ARM7opLDMIAuw     ,3},
    {ARM7opLDMIAuw     ,3},
    {ARM7opLDMIAuw     ,3},
    {ARM7opLDMIAuw     ,3},
    {ARM7opLDMIAuw     ,3},
    {ARM7opLDMIAuw     ,3},
    {ARM7opLDMIAuw     ,3},
    {ARM7opLDMIAuw     ,3},
    {ARM7opLDMIAuw     ,3},
    {ARM7opLDMIAuw     ,3},
    {ARM7opLDMIAuw     ,3},
    {ARM7opLDMIAuw     ,3},
    {ARM7opLDMIAuw     ,3},

    /*0x90*/
    {ARM7opSTMDB       ,2},
    {ARM7opSTMDB       ,2},
    {ARM7opSTMDB       ,2},
    {ARM7opSTMDB       ,2},
    {ARM7opSTMDB       ,2},
    {ARM7opSTMDB       ,2},
    {ARM7opSTMDB       ,2},
    {ARM7opSTMDB       ,2},
    {ARM7opSTMDB       ,2},
    {ARM7opSTMDB       ,2},
    {ARM7opSTMDB       ,2},
    {ARM7opSTMDB       ,2},
    {ARM7opSTMDB       ,2},
    {ARM7opSTMDB       ,2},
    {ARM7opSTMDB       ,2},
    {ARM7opSTMDB       ,2},

    /*0x91*/
    {ARM7opLDMDB       ,3},
    {ARM7opLDMDB       ,3},
    {ARM7opLDMDB       ,3},
    {ARM7opLDMDB       ,3},
    {ARM7opLDMDB       ,3},
    {ARM7opLDMDB       ,3},
    {ARM7opLDMDB       ,3},
    {ARM7opLDMDB       ,3},
    {ARM7opLDMDB       ,3},
    {ARM7opLDMDB       ,3},
    {ARM7opLDMDB       ,3},
    {ARM7opLDMDB       ,3},
    {ARM7opLDMDB       ,3},
    {ARM7opLDMDB       ,3},
    {ARM7opLDMDB       ,3},
    {ARM7opLDMDB       ,3},

    /*0x92*/
    {ARM7opSTMDBw      ,2},
    {ARM7opSTMDBw      ,2},
    {ARM7opSTMDBw      ,2},
    {ARM7opSTMDBw      ,2},
    {ARM7opSTMDBw      ,2},
    {ARM7opSTMDBw      ,2},
    {ARM7opSTMDBw      ,2},
    {ARM7opSTMDBw      ,2},
    {ARM7opSTMDBw      ,2},
    {ARM7opSTMDBw      ,2},
    {ARM7opSTMDBw      ,2},
    {ARM7opSTMDBw      ,2},
    {ARM7opSTMDBw      ,2},
    {ARM7opSTMDBw      ,2},
    {ARM7opSTMDBw      ,2},
    {ARM7opSTMDBw      ,2},

    /*0x93*/
    {ARM7opLDMDBw      ,3},
    {ARM7opLDMDBw      ,3},
    {ARM7opLDMDBw      ,3},
    {ARM7opLDMDBw      ,3},
    {ARM7opLDMDBw      ,3},
    {ARM7opLDMDBw      ,3},
    {ARM7opLDMDBw      ,3},
    {ARM7opLDMDBw      ,3},
    {ARM7opLDMDBw      ,3},
    {ARM7opLDMDBw      ,3},
    {ARM7opLDMDBw      ,3},
    {ARM7opLDMDBw      ,3},
    {ARM7opLDMDBw      ,3},
    {ARM7opLDMDBw      ,3},
    {ARM7opLDMDBw      ,3},
    {ARM7opLDMDBw      ,3},

    /*0x94*/
    {ARM7opSTMDBu      ,2},
    {ARM7opSTMDBu      ,2},
    {ARM7opSTMDBu      ,2},
    {ARM7opSTMDBu      ,2},
    {ARM7opSTMDBu      ,2},
    {ARM7opSTMDBu      ,2},
    {ARM7opSTMDBu      ,2},
    {ARM7opSTMDBu      ,2},
    {ARM7opSTMDBu      ,2},
    {ARM7opSTMDBu      ,2},
    {ARM7opSTMDBu      ,2},
    {ARM7opSTMDBu      ,2},
    {ARM7opSTMDBu      ,2},
    {ARM7opSTMDBu      ,2},
    {ARM7opSTMDBu      ,2},
    {ARM7opSTMDBu      ,2},

    /*0x95*/
    {ARM7opLDMDBu      ,3},
    {ARM7opLDMDBu      ,3},
    {ARM7opLDMDBu      ,3},
    {ARM7opLDMDBu      ,3},
    {ARM7opLDMDBu      ,3},
    {ARM7opLDMDBu      ,3},
    {ARM7opLDMDBu      ,3},
    {ARM7opLDMDBu      ,3},
    {ARM7opLDMDBu      ,3},
    {ARM7opLDMDBu      ,3},
    {ARM7opLDMDBu      ,3},
    {ARM7opLDMDBu      ,3},
    {ARM7opLDMDBu      ,3},
    {ARM7opLDMDBu      ,3},
    {ARM7opLDMDBu      ,3},
    {ARM7opLDMDBu      ,3},

    /*0x96*/
    {ARM7opSTMDBuw     ,2},
    {ARM7opSTMDBuw     ,2},
    {ARM7opSTMDBuw     ,2},
    {ARM7opSTMDBuw     ,2},
    {ARM7opSTMDBuw     ,2},
    {ARM7opSTMDBuw     ,2},
    {ARM7opSTMDBuw     ,2},
    {ARM7opSTMDBuw     ,2},
    {ARM7opSTMDBuw     ,2},
    {ARM7opSTMDBuw     ,2},
    {ARM7opSTMDBuw     ,2},
    {ARM7opSTMDBuw     ,2},
    {ARM7opSTMDBuw     ,2},
    {ARM7opSTMDBuw     ,2},
    {ARM7opSTMDBuw     ,2},
    {ARM7opSTMDBuw     ,2},

    /*0x97*/
    {ARM7opLDMDBuw     ,3},
    {ARM7opLDMDBuw     ,3},
    {ARM7opLDMDBuw     ,3},
    {ARM7opLDMDBuw     ,3},
    {ARM7opLDMDBuw     ,3},
    {ARM7opLDMDBuw     ,3},
    {ARM7opLDMDBuw     ,3},
    {ARM7opLDMDBuw     ,3},
    {ARM7opLDMDBuw     ,3},
    {ARM7opLDMDBuw     ,3},
    {ARM7opLDMDBuw     ,3},
    {ARM7opLDMDBuw     ,3},
    {ARM7opLDMDBuw     ,3},
    {ARM7opLDMDBuw     ,3},
    {ARM7opLDMDBuw     ,3},
    {ARM7opLDMDBuw     ,3},

    /*0x98*/
    {ARM7opSTMIB       ,2},
    {ARM7opSTMIB       ,2},
    {ARM7opSTMIB       ,2},
    {ARM7opSTMIB       ,2},
    {ARM7opSTMIB       ,2},
    {ARM7opSTMIB       ,2},
    {ARM7opSTMIB       ,2},
    {ARM7opSTMIB       ,2},
    {ARM7opSTMIB       ,2},
    {ARM7opSTMIB       ,2},
    {ARM7opSTMIB       ,2},
    {ARM7opSTMIB       ,2},
    {ARM7opSTMIB       ,2},
    {ARM7opSTMIB       ,2},
    {ARM7opSTMIB       ,2},
    {ARM7opSTMIB       ,2},

    /*0x99*/
    {ARM7opLDMIB       ,3},
    {ARM7opLDMIB       ,3},
    {ARM7opLDMIB       ,3},
    {ARM7opLDMIB       ,3},
    {ARM7opLDMIB       ,3},
    {ARM7opLDMIB       ,3},
    {ARM7opLDMIB       ,3},
    {ARM7opLDMIB       ,3},
    {ARM7opLDMIB       ,3},
    {ARM7opLDMIB       ,3},
    {ARM7opLDMIB       ,3},
    {ARM7opLDMIB       ,3},
    {ARM7opLDMIB       ,3},
    {ARM7opLDMIB       ,3},
    {ARM7opLDMIB       ,3},
    {ARM7opLDMIB       ,3},

    /*0x9A*/
    {ARM7opSTMIBw      ,2},
    {ARM7opSTMIBw      ,2},
    {ARM7opSTMIBw      ,2},
    {ARM7opSTMIBw      ,2},
    {ARM7opSTMIBw      ,2},
    {ARM7opSTMIBw      ,2},
    {ARM7opSTMIBw      ,2},
    {ARM7opSTMIBw      ,2},
    {ARM7opSTMIBw      ,2},
    {ARM7opSTMIBw      ,2},
    {ARM7opSTMIBw      ,2},
    {ARM7opSTMIBw      ,2},
    {ARM7opSTMIBw      ,2},
    {ARM7opSTMIBw      ,2},
    {ARM7opSTMIBw      ,2},
    {ARM7opSTMIBw      ,2},

    /*0x9B*/
    {ARM7opLDMIBw      ,3},
    {ARM7opLDMIBw      ,3},
    {ARM7opLDMIBw      ,3},
    {ARM7opLDMIBw      ,3},
    {ARM7opLDMIBw      ,3},
    {ARM7opLDMIBw      ,3},
    {ARM7opLDMIBw      ,3},
    {ARM7opLDMIBw      ,3},
    {ARM7opLDMIBw      ,3},
    {ARM7opLDMIBw      ,3},
    {ARM7opLDMIBw      ,3},
    {ARM7opLDMIBw      ,3},
    {ARM7opLDMIBw      ,3},
    {ARM7opLDMIBw      ,3},
    {ARM7opLDMIBw      ,3},
    {ARM7opLDMIBw      ,3},

    /*0x9C*/
    {ARM7opSTMIBu      ,2},
    {ARM7opSTMIBu      ,2},
    {ARM7opSTMIBu      ,2},
    {ARM7opSTMIBu      ,2},
    {ARM7opSTMIBu      ,2},
    {ARM7opSTMIBu      ,2},
    {ARM7opSTMIBu      ,2},
    {ARM7opSTMIBu      ,2},
    {ARM7opSTMIBu      ,2},
    {ARM7opSTMIBu      ,2},
    {ARM7opSTMIBu      ,2},
    {ARM7opSTMIBu      ,2},
    {ARM7opSTMIBu      ,2},
    {ARM7opSTMIBu      ,2},
    {ARM7opSTMIBu      ,2},
    {ARM7opSTMIBu      ,2},

    /*0x9D*/
    {ARM7opLDMIBu      ,3},
    {ARM7opLDMIBu      ,3},
    {ARM7opLDMIBu      ,3},
    {ARM7opLDMIBu      ,3},
    {ARM7opLDMIBu      ,3},
    {ARM7opLDMIBu      ,3},
    {ARM7opLDMIBu      ,3},
    {ARM7opLDMIBu      ,3},
    {ARM7opLDMIBu      ,3},
    {ARM7opLDMIBu      ,3},
    {ARM7opLDMIBu      ,3},
    {ARM7opLDMIBu      ,3},
    {ARM7opLDMIBu      ,3},
    {ARM7opLDMIBu      ,3},
    {ARM7opLDMIBu      ,3},
    {ARM7opLDMIBu      ,3},

    /*0x9E*/
    {ARM7opSTMIBuw     ,2},
    {ARM7opSTMIBuw     ,2},
    {ARM7opSTMIBuw     ,2},
    {ARM7opSTMIBuw     ,2},
    {ARM7opSTMIBuw     ,2},
    {ARM7opSTMIBuw     ,2},
    {ARM7opSTMIBuw     ,2},
    {ARM7opSTMIBuw     ,2},
    {ARM7opSTMIBuw     ,2},
    {ARM7opSTMIBuw     ,2},
    {ARM7opSTMIBuw     ,2},
    {ARM7opSTMIBuw     ,2},
    {ARM7opSTMIBuw     ,2},
    {ARM7opSTMIBuw     ,2},
    {ARM7opSTMIBuw     ,2},
    {ARM7opSTMIBuw     ,2},

    /*0x9F*/
    {ARM7opLDMIBuw     ,3},
    {ARM7opLDMIBuw     ,3},
    {ARM7opLDMIBuw     ,3},
    {ARM7opLDMIBuw     ,3},
    {ARM7opLDMIBuw     ,3},
    {ARM7opLDMIBuw     ,3},
    {ARM7opLDMIBuw     ,3},
    {ARM7opLDMIBuw     ,3},
    {ARM7opLDMIBuw     ,3},
    {ARM7opLDMIBuw     ,3},
    {ARM7opLDMIBuw     ,3},
    {ARM7opLDMIBuw     ,3},
    {ARM7opLDMIBuw     ,3},
    {ARM7opLDMIBuw     ,3},
    {ARM7opLDMIBuw     ,3},
    {ARM7opLDMIBuw     ,3},

    /*0xA0*/
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},

    /*0xA1*/
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},

    /*0xA2*/
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},

    /*0xA3*/
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},

    /*0xA4*/
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},

    /*0xA5*/
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},

    /*0xA6*/
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},

    /*0xA7*/
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},

    /*0xA8*/
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},

    /*0xA9*/
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},

    /*0xAA*/
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},

    /*0xAB*/
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},

    /*0xAC*/
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},

    /*0xAD*/
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},

    /*0xAE*/
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},

    /*0xAF*/
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},
    {ARM7opB           ,3},

    /*0xB0*/
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},

    /*0xB1*/
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},

    /*0xB2*/
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},

    /*0xB3*/
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},

    /*0xB4*/
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},

    /*0xB5*/
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},

    /*0xB6*/
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},

    /*0xB7*/
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},

    /*0xB8*/
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},

    /*0xB9*/
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},

    /*0xBA*/
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},

    /*0xBB*/
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},

    /*0xBC*/
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},

    /*0xBD*/
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},

    /*0xBE*/
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},

    /*0xBF*/
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},
    {ARM7opBL          ,3},

    /*0xC0*/ // Coprocessors undefined here!
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xC1*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xC2*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xC3*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xC4*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xC5*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xC6*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xC7*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xC8*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xC9*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xCA*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xCB*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xCC*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xCD*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xCE*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xCF*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xD0*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xD1*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xD2*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xD3*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xD4*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xD5*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xD6*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xD7*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xD8*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xD9*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xDA*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xDB*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xDC*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xDD*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xDE*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xDF*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xE0*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xE1*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xE2*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xE3*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xE4*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xE5*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xE6*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xE7*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xE8*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xE9*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xEA*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xEB*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xEC*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xED*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xEE*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xEF*/
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},
    {ARM7opUNI         ,1},

    /*0xF0*/
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},

    /*0xF1*/
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},

    /*0xF2*/
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},

    /*0xF3*/
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},

    /*0xF4*/
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},

    /*0xF5*/
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},

    /*0xF6*/
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},

    /*0xF7*/
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},

    /*0xF8*/
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},

    /*0xF9*/
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},

    /*0xFA*/
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},

    /*0xFB*/
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},

    /*0xFC*/
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},

    /*0xFD*/
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},

    /*0xFE*/
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},

    /*0xFF*/
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
    {ARM7opSWI         ,3},
};

const ARM7OPDESC thumb7ops[]={

    //0x00
    {Thumb7opLSLimm    ,1},
    {Thumb7opLSLimm    ,1},
    {Thumb7opLSLimm    ,1},
    {Thumb7opLSLimm    ,1},
    {Thumb7opLSLimm    ,1},
    {Thumb7opLSLimm    ,1},
    {Thumb7opLSLimm    ,1},
    {Thumb7opLSLimm    ,1},
    {Thumb7opLSRimm    ,1},
    {Thumb7opLSRimm    ,1},
    {Thumb7opLSRimm    ,1},
    {Thumb7opLSRimm    ,1},
    {Thumb7opLSRimm    ,1},
    {Thumb7opLSRimm    ,1},
    {Thumb7opLSRimm    ,1},
    {Thumb7opLSRimm    ,1},

    //0x10
    {Thumb7opASRimm    ,1},
    {Thumb7opASRimm    ,1},
    {Thumb7opASRimm    ,1},
    {Thumb7opASRimm    ,1},
    {Thumb7opASRimm    ,1},
    {Thumb7opASRimm    ,1},
    {Thumb7opASRimm    ,1},
    {Thumb7opASRimm    ,1},
    {Thumb7opADDreg    ,1},
    {Thumb7opADDreg    ,1},
    {Thumb7opSUBreg    ,1},
    {Thumb7opSUBreg    ,1},
    {Thumb7opADDimm3   ,1},
    {Thumb7opADDimm3   ,1},
    {Thumb7opSUBimm3   ,1},
    {Thumb7opSUBimm3   ,1},

    //0x20
    {Thumb7opMOVimm8r0 ,1},
    {Thumb7opMOVimm8r1 ,1},
    {Thumb7opMOVimm8r2 ,1},
    {Thumb7opMOVimm8r3 ,1},
    {Thumb7opMOVimm8r4 ,1},
    {Thumb7opMOVimm8r5 ,1},
    {Thumb7opMOVimm8r6 ,1},
    {Thumb7opMOVimm8r7 ,1},
    {Thumb7opCMPimm8r0 ,1},
    {Thumb7opCMPimm8r1 ,1},
    {Thumb7opCMPimm8r2 ,1},
    {Thumb7opCMPimm8r3 ,1},
    {Thumb7opCMPimm8r4 ,1},
    {Thumb7opCMPimm8r5 ,1},
    {Thumb7opCMPimm8r6 ,1},
    {Thumb7opCMPimm8r7 ,1},

    //0x30
    {Thumb7opADDimm8r0 ,1},
    {Thumb7opADDimm8r1 ,1},
    {Thumb7opADDimm8r2 ,1},
    {Thumb7opADDimm8r3 ,1},
    {Thumb7opADDimm8r4 ,1},
    {Thumb7opADDimm8r5 ,1},
    {Thumb7opADDimm8r6 ,1},
    {Thumb7opADDimm8r7 ,1},
    {Thumb7opSUBimm8r0 ,1},
    {Thumb7opSUBimm8r1 ,1},
    {Thumb7opSUBimm8r2 ,1},
    {Thumb7opSUBimm8r3 ,1},
    {Thumb7opSUBimm8r4 ,1},
    {Thumb7opSUBimm8r5 ,1},
    {Thumb7opSUBimm8r6 ,1},
    {Thumb7opSUBimm8r7 ,1},

    //0x40
    {Thumb7opDPg1      ,1},
    {Thumb7opDPg2      ,1},
    {Thumb7opDPg3      ,1},
    {Thumb7opDPg4      ,1},
    {Thumb7opADDH      ,1},
    {Thumb7opCMPH      ,1},
    {Thumb7opMOVH      ,1},
    {Thumb7opBXreg     ,3},
    {Thumb7opLDRPCr0   ,3},
    {Thumb7opLDRPCr1   ,3},
    {Thumb7opLDRPCr2   ,3},
    {Thumb7opLDRPCr3   ,3},
    {Thumb7opLDRPCr4   ,3},
    {Thumb7opLDRPCr5   ,3},
    {Thumb7opLDRPCr6   ,3},
    {Thumb7opLDRPCr7   ,3},

    //0x50
    {Thumb7opSTRreg    ,2},
    {Thumb7opSTRreg    ,2},
    {Thumb7opSTRHreg   ,2},
    {Thumb7opSTRHreg   ,2},
    {Thumb7opSTRBreg   ,2},
    {Thumb7opSTRBreg   ,2},
    {Thumb7opLDRSBreg  ,3},
    {Thumb7opLDRSBreg  ,3},
    {Thumb7opLDRreg    ,3},
    {Thumb7opLDRreg    ,3},
    {Thumb7opLDRHreg   ,3},
    {Thumb7opLDRHreg   ,3},
    {Thumb7opLDRBreg   ,3},
    {Thumb7opLDRBreg   ,3},
    {Thumb7opLDRSHreg  ,3},
    {Thumb7opLDRSHreg  ,3},

    //0x60
    {Thumb7opSTRimm5   ,2},
    {Thumb7opSTRimm5   ,2},
    {Thumb7opSTRimm5   ,2},
    {Thumb7opSTRimm5   ,2},
    {Thumb7opSTRimm5   ,2},
    {Thumb7opSTRimm5   ,2},
    {Thumb7opSTRimm5   ,2},
    {Thumb7opSTRimm5   ,2},
    {Thumb7opLDRimm5   ,3},
    {Thumb7opLDRimm5   ,3},
    {Thumb7opLDRimm5   ,3},
    {Thumb7opLDRimm5   ,3},
    {Thumb7opLDRimm5   ,3},
    {Thumb7opLDRimm5   ,3},
    {Thumb7opLDRimm5   ,3},
    {Thumb7opLDRimm5   ,3},

    //0x70
    {Thumb7opSTRBimm5  ,2},
    {Thumb7opSTRBimm5  ,2},
    {Thumb7opSTRBimm5  ,2},
    {Thumb7opSTRBimm5  ,2},
    {Thumb7opSTRBimm5  ,2},
    {Thumb7opSTRBimm5  ,2},
    {Thumb7opSTRBimm5  ,2},
    {Thumb7opSTRBimm5  ,2},
    {Thumb7opLDRBimm5  ,3},
    {Thumb7opLDRBimm5  ,3},
    {Thumb7opLDRBimm5  ,3},
    {Thumb7opLDRBimm5  ,3},
    {Thumb7opLDRBimm5  ,3},
    {Thumb7opLDRBimm5  ,3},
    {Thumb7opLDRBimm5  ,3},
    {Thumb7opLDRBimm5  ,3},

    //0x80
    {Thumb7opSTRHimm5  ,2},
    {Thumb7opSTRHimm5  ,2},
    {Thumb7opSTRHimm5  ,2},
    {Thumb7opSTRHimm5  ,2},
    {Thumb7opSTRHimm5  ,2},
    {Thumb7opSTRHimm5  ,2},
    {Thumb7opSTRHimm5  ,2},
    {Thumb7opSTRHimm5  ,2},
    {Thumb7opLDRHimm5  ,3},
    {Thumb7opLDRHimm5  ,3},
    {Thumb7opLDRHimm5  ,3},
    {Thumb7opLDRHimm5  ,3},
    {Thumb7opLDRHimm5  ,3},
    {Thumb7opLDRHimm5  ,3},
    {Thumb7opLDRHimm5  ,3},
    {Thumb7opLDRHimm5  ,3},

    //0x90
    {Thumb7opSTRSPr0   ,2},
    {Thumb7opSTRSPr1   ,2},
    {Thumb7opSTRSPr2   ,2},
    {Thumb7opSTRSPr3   ,2},
    {Thumb7opSTRSPr4   ,2},
    {Thumb7opSTRSPr5   ,2},
    {Thumb7opSTRSPr6   ,2},
    {Thumb7opSTRSPr7   ,2},
    {Thumb7opLDRSPr0   ,3},
    {Thumb7opLDRSPr1   ,3},
    {Thumb7opLDRSPr2   ,3},
    {Thumb7opLDRSPr3   ,3},
    {Thumb7opLDRSPr4   ,3},
    {Thumb7opLDRSPr5   ,3},
    {Thumb7opLDRSPr6   ,3},
    {Thumb7opLDRSPr7   ,3},

    //0xA0
    {Thumb7opADDPCr0   ,1},
    {Thumb7opADDPCr1   ,1},
    {Thumb7opADDPCr2   ,1},
    {Thumb7opADDPCr3   ,1},
    {Thumb7opADDPCr4   ,1},
    {Thumb7opADDPCr5   ,1},
    {Thumb7opADDPCr6   ,1},
    {Thumb7opADDPCr7   ,1},
    {Thumb7opADDSPr0   ,1},
    {Thumb7opADDSPr1   ,1},
    {Thumb7opADDSPr2   ,1},
    {Thumb7opADDSPr3   ,1},
    {Thumb7opADDSPr4   ,1},
    {Thumb7opADDSPr5   ,1},
    {Thumb7opADDSPr6   ,1},
    {Thumb7opADDSPr7   ,1},

    //0xB0
    {Thumb7opADDSPimm7 ,1},
    {Thumb7opUND       ,1},  // Misc map: unallocated
    {Thumb7opUND       ,1},  // Misc map: unallocated
    {Thumb7opUND       ,1},  // Misc map: unallocated
    {Thumb7opPUSH      ,1},
    {Thumb7opPUSHlr    ,1},
    {Thumb7opUND       ,1},  // Misc map: unallocated
    {Thumb7opUND       ,1},  // Misc map: unallocated
    {Thumb7opUND       ,1},  // Misc map: unallocated
    {Thumb7opUND       ,1},  // Misc map: unallocated
    {Thumb7opUND       ,1},  // Misc map: unallocated
    {Thumb7opUND       ,1},  // Misc map: unallocated
    {Thumb7opPOP       ,1},
    {Thumb7opPOPpc     ,1},
    {Thumb7opBKPT      ,1},
    {Thumb7opUND       ,1},  // Misc map: unallocated

    //0xC0
    {Thumb7opSTMIAr0   ,1},
    {Thumb7opSTMIAr1   ,1},
    {Thumb7opSTMIAr2   ,1},
    {Thumb7opSTMIAr3   ,1},
    {Thumb7opSTMIAr4   ,1},
    {Thumb7opSTMIAr5   ,1},
    {Thumb7opSTMIAr6   ,1},
    {Thumb7opSTMIAr7   ,1},
    {Thumb7opLDMIAr0   ,1},
    {Thumb7opLDMIAr1   ,1},
    {Thumb7opLDMIAr2   ,1},
    {Thumb7opLDMIAr3   ,1},
    {Thumb7opLDMIAr4   ,1},
    {Thumb7opLDMIAr5   ,1},
    {Thumb7opLDMIAr6   ,1},
    {Thumb7opLDMIAr7   ,1},

    //0xD0
    {Thumb7opBEQ       ,1},
    {Thumb7opBNE       ,1},
    {Thumb7opBCS       ,1},
    {Thumb7opBCC       ,1},
    {Thumb7opBMI       ,1},
    {Thumb7opBPL       ,1},
    {Thumb7opBVS       ,1},
    {Thumb7opBVC       ,1},
    {Thumb7opBHI       ,1},
    {Thumb7opBLS       ,1},
    {Thumb7opBGE       ,1},
    {Thumb7opBLT       ,1},
    {Thumb7opBGT       ,1},
    {Thumb7opBLE       ,1},
    {Thumb7opUND       ,1},  // Covered by unconditional
    {Thumb7opSWI       ,3},

    //0xE0
    {Thumb7opB         ,3},
    {Thumb7opB         ,3},
    {Thumb7opB         ,3},
    {Thumb7opB         ,3},
    {Thumb7opB         ,3},
    {Thumb7opB         ,3},
    {Thumb7opB         ,3},
    {Thumb7opB         ,3},
    {Thumb7opUND       ,1},  // Defined in ARM5!
    {Thumb7opUND       ,1},  // Defined in ARM5!
    {Thumb7opUND       ,1},  // Defined in ARM5!
    {Thumb7opUND       ,1},  // Defined in ARM5!
    {Thumb7opUND       ,1},  // Defined in ARM5!
    {Thumb7opUND       ,1},  // Defined in ARM5!
    {Thumb7opUND       ,1},  // Defined in ARM5!
    {Thumb7opUND       ,1},  // Defined in ARM5!

    //0xF0
    {Thumb7opBLsetup   ,4},
    {Thumb7opBLsetup   ,4},
    {Thumb7opBLsetup   ,4},
    {Thumb7opBLsetup   ,4},
    {Thumb7opBLsetup   ,4},
    {Thumb7opBLsetup   ,4},
    {Thumb7opBLsetup   ,4},
    {Thumb7opBLsetup   ,4},
    {Thumb7opBLoff     ,4},
    {Thumb7opBLoff     ,4},
    {Thumb7opBLoff     ,4},
    {Thumb7opBLoff     ,4},
    {Thumb7opBLoff     ,4},
    {Thumb7opBLoff     ,4},
    {Thumb7opBLoff     ,4},
    {Thumb7opBLoff     ,4},
};

#endif//__ARM7FUNC_H_

/*** EOF:arm7func.h ******************************************************/

